from abc import ABC, abstractmethod
from typing import Optional, Tuple


class PkgRegistry(ABC):
    @abstractmethod
    def get_latest_version_and_hash(self, pkg_name: str) -> Optional[Tuple[str, str]]:
        """Get the latest version of the package and its hash

        Args:
            pkg_name: Name of the package

        Returns:
            Tuple[str, str]: (version, hash)
        """
        raise NotImplementedError()



from __future__ import annotations
from dataclasses import dataclass
from distutils.version import Version
from enum import Enum
from pathlib import Path
from typing import Dict, List, Optional
from typing_extensions import TypeGuard

from semver import VersionInfo
from loguru import logger


@dataclass
class Package:
    name: str
    version: str
    dependencies: Dict[str, DepConstraints]
    dev_dependencies: Dict[str, DepConstraints]

    # below are properties that can be from the package file, but may be modified heavily
    # so that you should never use them to override previous package definition
    type: PackageType
    location: Path
    # list of glob patterns to be included in the final package
    include: List[str]
    # a list of glob patterns to be excluded in the final package
    exclude: List[str]

    def get_all_dependency_names(self) -> List[str]:
        out = list(self.dependencies.keys())
        out.extend(self.dev_dependencies.keys())
        return out


class PackageType(str, Enum):
    Poetry = "poetry"
    Maturin = "maturin"

    def is_compatible(self, other: PackageType) -> bool:
        return (
            self == other
            or (self == PackageType.Poetry and other == PackageType.Maturin)
            or (self == PackageType.Maturin and other == PackageType.Poetry)
        )


@dataclass(eq=True)
class DepConstraint:
    """Constraint of a dependency.

    The two important fields are rule (for comparing between versions) and constraint
    (for distinguish between different platforms/os).

    To reconstruct/save the constraint,
    back to the original package specification, update the `origin_spec` with the a new key
    stored in `version_field` and value from `version`. The reason for this behaviour is to
    support cases such as where the dependency is from git (`version_field` = 'git').
    """

    # rule for matching version of dependency, e.g. "^1.0.0" or ">= 1.0.0", the rule sometimes depends on what package's type
    version_spec: str
    # an identifier for the condition that this version is applicable to.
    # none mean there is no other constraint.
    constraint: Optional[str] = None
    # name of the rule field in origin specification
    # none if the spec is just a string
    version_spec_field: Optional[str] = None
    # the original specification without the version
    # none if the spec is just a string
    origin_spec: Optional[dict] = None


# see: https://python-poetry.org/docs/dependency-specification/
# the constraints always sorted by constraint
DepConstraints = List[DepConstraint]


@dataclass
class VersionSpec:
    lowerbound: Optional[VersionInfo]
    upperbound: Optional[VersionInfo]
    is_lowerbound_inclusive: bool
    is_upperbound_inclusive: bool

    def is_version_compatible(self, version: VersionInfo) -> bool:
        """Check if the given version is compatible with the given rule

        Args:
            version: the version to check
        """
        incompatible = (
            self.lowerbound is not None
            and (
                (self.is_lowerbound_inclusive and self.lowerbound > version)
                or (not self.is_lowerbound_inclusive and self.lowerbound >= version)
            )
        ) or (
            self.upperbound is not None
            and (
                (self.is_upperbound_inclusive and self.upperbound < version)
                or (not self.is_upperbound_inclusive and self.upperbound <= version)
            )
        )
        return not incompatible

    def intersect(self, version_spec: VersionSpec) -> VersionSpec:
        """Intersect two version specs. Result in a stricter version spec.

        Raise exception if the intersection is empty.

        Examples:
            - "^1.0.0" and "^2.0.0" -> Exception
            - "^1.1.0" and "^1.2.0" -> "^1.2.0"
        """
        lb = self.lowerbound
        is_lb_inclusive = self.is_lowerbound_inclusive

        if version_spec.lowerbound is not None:
            if lb is None:
                lb = version_spec.lowerbound
                is_lb_inclusive = version_spec.is_lowerbound_inclusive
            elif version_spec.lowerbound > lb:
                lb = version_spec.lowerbound
                is_lb_inclusive = version_spec.is_lowerbound_inclusive
            elif version_spec.lowerbound == lb:
                is_lb_inclusive = (
                    is_lb_inclusive and version_spec.is_lowerbound_inclusive
                )

        ub = self.upperbound
        is_ub_inclusive = self.is_upperbound_inclusive

        if version_spec.upperbound is not None:
            if ub is None:
                ub = version_spec.upperbound
                is_ub_inclusive = version_spec.is_upperbound_inclusive
            elif version_spec.upperbound < ub:
                ub = version_spec.upperbound
                is_ub_inclusive = version_spec.is_upperbound_inclusive
            elif version_spec.upperbound == ub:
                is_ub_inclusive = (
                    is_ub_inclusive and version_spec.is_upperbound_inclusive
                )

        if lb is not None and ub is not None and lb > ub:
            raise ValueError(
                "Can't intersect two version specs: {} and {} because it results in empty spec".format(
                    self, version_spec
                )
            )

        return VersionSpec(
            lowerbound=lb,
            upperbound=ub,
            is_lowerbound_inclusive=is_lb_inclusive,
            is_upperbound_inclusive=is_ub_inclusive,
        )

    def to_pep508_string(self):
        s = f">{'=' if self.is_lowerbound_inclusive else ''} {str(self.lowerbound)}"
        if self.upperbound is not None:
            s += f", <{'=' if self.is_upperbound_inclusive else ''} {str(self.upperbound)}"
        return s

    def __eq__(self, other: VersionSpec):
        if other is None or not isinstance(other, VersionSpec):
            return False

        return (
            (
                (
                    self.lowerbound is not None
                    and other.lowerbound is not None
                    and self.lowerbound == other.lowerbound
                )
                or (self.lowerbound is None and other.lowerbound is None)
            )
            and (
                (
                    self.upperbound is not None
                    and other.upperbound is not None
                    and self.upperbound == other.upperbound
                )
                or (self.upperbound is None and other.upperbound is None)
            )
            and (self.is_lowerbound_inclusive == other.is_lowerbound_inclusive)
            and (self.is_upperbound_inclusive == other.is_upperbound_inclusive)
        )



import enum
from operator import itemgetter
from typing import Dict, List, Optional, Set

from loguru import logger
from pbt.config import PBTConfig
from pbt.diff import RemoteDiff
from pbt.package.graph import PkgGraph, ThirdPartyPackage
from pbt.package.manager.manager import PkgManager, build_cache
from pbt.package.package import (
    Package,
    PackageType,
    DepConstraint,
)
from pbt.package.registry.registry import PkgRegistry
from loguru import logger


class VersionConsistent(enum.Enum):
    # it is consistent if it is exact match
    STRICT = "strict"
    # it is consistent if it matches the constraint
    COMPATIBLE = "compatible"
    # no consistent check is performed, and the latest version is preferred
    NO_CHECK = "no_check"


class BTPipeline:
    def __init__(self, cfg: PBTConfig, managers: Dict[PackageType, PkgManager]) -> None:
        self.root = cfg.cwd
        self.cfg = cfg
        self.managers = managers
        self.graph = PkgGraph()
        self.pkgs: Dict[str, Package] = {}

    def discover(self):
        """Discover packages in the project."""
        logger.debug("Discovering packages in the project...")
        pkgs = {}
        for manager in self.managers.values():
            for fpath in manager.discover(
                self.root, self.cfg.ignore_directories, self.cfg.ignore_directory_names
            ):
                if not manager.is_package_directory(fpath):
                    continue
                pkg = manager.load(fpath)
                if pkg.name in pkgs:
                    raise RuntimeError(f"Duplicate package {pkg.name}")
                pkgs[pkg.name] = pkg
        self.graph = PkgGraph.from_pkgs(pkgs)
        self.pkgs = pkgs
        logger.debug("Discovering packages in the project... Done!")

    def enforce_version_consistency(
        self,
        mode: VersionConsistent = VersionConsistent.COMPATIBLE,
        thirdparty_mode: VersionConsistent = VersionConsistent.COMPATIBLE,
        freeze_packages: Optional[Set[str]] = None,
    ):
        """Update version of packages & third-party packages in the project.

        Args:
            mode: how to enforce version consistency for your OWN packages in the project.
            thirdparty_mode: how to enforce version consistency for third-party packages in the project.
        """
        # resolve the latest version specs of (third-party) packages
        thirdparty_pkgs = {}
        freeze_packages = freeze_packages or set()

        for pkg in self.graph.iter_pkg():
            if isinstance(pkg, ThirdPartyPackage):
                manager = self.managers[pkg.type]

                try:
                    pkg_spec = manager.find_latest_specs(
                        [
                            v
                            for k, v in pkg.invert_dependencies.items()
                            if k not in freeze_packages
                        ],
                        mode="strict"
                        if thirdparty_mode == VersionConsistent.STRICT
                        else (
                            "compatible"
                            if thirdparty_mode == VersionConsistent.COMPATIBLE
                            else None
                        ),
                    )
                except ValueError as e:
                    raise ValueError(
                        f"Package {pkg.name} has different versions used in those packages: {list(pkg.invert_dependencies.keys())}. Consider fixing it"
                    ) from e

                thirdparty_pkgs[pkg.name] = pkg_spec

                # update graph
                for key in pkg.invert_dependencies:
                    if key not in freeze_packages:
                        pkg.invert_dependencies[key] = pkg_spec

        # iterate over packages and update their dependencies.
        # however, for your own packages, always use the latest version or make sure it
        # is compatible according to the mode
        for pkg in self.pkgs.values():
            if pkg.name in freeze_packages:
                continue
            manager = self.managers[pkg.type]
            is_modified = False
            for deps in [pkg.dependencies, pkg.dev_dependencies]:
                for dep, specs in deps.items():
                    if dep in manager.get_fixed_version_pkgs():
                        # skip fixed version packages
                        continue

                    if dep in self.pkgs:
                        # update your own packages
                        dep_version = manager.parse_version(self.pkgs[dep].version)
                        if mode == VersionConsistent.COMPATIBLE:
                            for spec in specs:
                                if not manager.is_version_compatible(
                                    dep_version, spec.version_spec
                                ):
                                    spec.version_spec = manager.update_version_spec(
                                        spec.version_spec, dep_version
                                    )
                                    is_modified = True
                        else:
                            assert mode == VersionConsistent.STRICT
                            for spec in specs:
                                version_spec = manager.parse_version_spec(
                                    spec.version_spec
                                )
                                if version_spec.lowerbound != dep_version:
                                    is_modified = True
                                    spec.version_spec = manager.update_version_spec(
                                        spec.version_spec, dep_version
                                    )
                    elif specs != thirdparty_pkgs[dep]:
                        deps[dep] = thirdparty_pkgs[dep]
                        is_modified = True

            if is_modified:
                manager.save(pkg)

    def install(
        self,
        pkg_names: List[str],
        include_dev: bool = False,
    ):
        """Install packages

        Args:
            pkg_names: name of packages to install
            include_dev: whether to install development dependencies
        """
        pkgs = [self.pkgs[name] for name in pkg_names]

        with build_cache():
            for pkg in pkgs:
                manager = self.managers[pkg.type]
                # gather all dependencies in one file and install it.
                deps = self.graph.dependencies(pkg.name, include_dev=include_dev)

                skip_deps = [
                    dep.name
                    for dep in deps
                    if isinstance(dep, Package)
                    and (dep.name not in self.cfg.use_prebuilt_binaries)
                ]
                additional_deps = {
                    dep.name: next(iter(dep.invert_dependencies.values()))
                    if isinstance(dep, ThirdPartyPackage)
                    else [DepConstraint(version_spec=dep.version)]
                    for dep in deps
                    if (
                        isinstance(dep, ThirdPartyPackage)
                        and dep.name not in pkg.dependencies
                        and dep.name not in pkg.dev_dependencies
                    )
                    or (
                        isinstance(dep, Package)
                        and dep.name not in pkg.dependencies
                        and dep.name not in pkg.dev_dependencies
                        and dep.name in self.cfg.use_prebuilt_binaries
                    )
                }

                logger.info("Installing package: {}", pkg.name)
                manager.install(
                    pkg,
                    include_dev=include_dev,
                    skip_deps=skip_deps,
                    additional_deps=additional_deps,
                )

                for dep in deps:
                    if (
                        isinstance(dep, Package)
                        and dep.name not in self.cfg.use_prebuilt_binaries
                    ):
                        logger.info("Installing local dependency: {}", dep.name)
                        manager.install_dependency(
                            pkg, dep, skip_dep_deps=dep.get_all_dependency_names()
                        )

    def publish(self, pkg_names: List[str], registries: Dict[PackageType, PkgRegistry]):
        """Publish a package. Check if the package is modified but the version is not changed so
        that we don't forget to update the version of the package.

        Args:
            pkg_names: name of the package to publish
            registries: registries to publish to
        """
        pkgs = [self.pkgs[name] for name in pkg_names]

        with build_cache():
            publishing_pkgs = {}

            for pkg in pkgs:
                publishing_pkgs[pkg.name] = pkg
                for dep in self.graph.dependencies(pkg.name, include_dev=False):
                    if isinstance(dep, Package):
                        publishing_pkgs[dep.name] = dep

            diffs = {}

            has_error = False
            for pkg in publishing_pkgs.values():
                remote_pkg_version, remote_pkg_hash = registries[
                    pkg.type
                ].get_latest_version_and_hash(pkg.name) or (None, None)
                diff = RemoteDiff.from_pkg(
                    self.managers[pkg.type], pkg, remote_pkg_version, remote_pkg_hash
                )
                if not diff.is_version_diff and diff.is_content_changed:
                    logger.error(
                        "Package {} has been modified, but its version hasn't been updated",
                        pkg.name,
                    )
                    has_error = True
                diffs[pkg.name] = diff
            if has_error:
                raise Exception(
                    "Stop publishing because some packages have been modified but their versions haven't been updated. Please see the logs for more information"
                )

            for name, pkg in sorted(publishing_pkgs.items(), key=itemgetter(0)):
                if diffs[name].is_version_diff:
                    self.managers[pkg.type].publish(pkg)



import functools
import os
from pathlib import Path
import subprocess
from typing import Callable, List, Sequence, Union, Optional
from typing_extensions import TypedDict


class ExecProcessError(Exception):
    pass


NewEnvVar = TypedDict("NewEnvVar", name=str, value=str)


def stdout(line):
    """Print line to stdout"""
    print(line)


def exec(
    cmd: Union[str, Sequence[Union[str, Path]]],
    handler: Optional[Callable[[str], None]] = None,
    check_returncode: bool = True,
    cwd: Union[Path, str] = "./",
    redirect_stderr: bool = False,
    capture_stdout: bool = True,
    env: Optional[Union[List[str], List[Union[str, NewEnvVar]], dict]] = None,
) -> List[str]:
    """
    Execute a command and return the list of lines, in which the newline character is stripped away.

    Args:
        cmd: Command to execute.
        handler: function to process each line of the output.
        check_returncode: Whether to check the return code.
        cwd: working directory.
        redirect_stderr: Whether to redirect stderr to stdout.
        capture_stdout: Whether to capture stdout. For running poetry command, we may not want to capture because they manipulate output lines.
        env: the environment variables to use in this process.
            - None is use the default behavior of Popen
            - a list of strings/dictionaries:
                - if the item is a string, it is the environment variable to pass from the parent process
                - if the item is a dictionary, it is the new environment variable to set, has the following format: { name: <name>, value: <value> }
            - a dictionary will be the environment variables to use
    """
    if isinstance(cmd, str):
        cmd = [x for x in cmd.split(" ") if x != ""]
    else:
        cmd = [str(x) for x in cmd]

    if handler is None:
        handler = lambda x: None

    if env is not None:
        if isinstance(env, list):
            tmp = {}
            for item in env:
                if isinstance(item, str):
                    if item in os.environ:
                        tmp[item] = os.environ[item]
                else:
                    assert isinstance(item, dict)
                    tmp[item["name"]] = item["value"]
            env = tmp

    if capture_stdout:
        p = subprocess.Popen(
            cmd,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT if redirect_stderr else None,
            cwd=str(cwd),
            env=env,
        )
        output = []

        while True:
            assert p.stdout is not None
            line = p.stdout.readline().decode("utf-8")
            if line != "":
                assert line[-1] == "\n"
                line = line[:-1]
                output.append(line)
                handler(line)
            elif p.poll() is not None:
                break

        returncode = p.returncode
    else:
        returncode = subprocess.call(cmd, cwd=str(cwd), env=env)
        output = []

    if check_returncode and returncode != 0:
        msg = (
            f"Command: f{cmd} returns non-zero exit status {returncode}\n"
            + "\n".join(output)
        )
        raise ExecProcessError(msg)

    return output


def cache_func():
    d = {}

    def wrapper(func):
        @functools.wraps(func)
        def fn(*args):
            if args not in d:
                d[args] = func(*args)
            return d[args]

        return fn

    return wrapper


def cache_method():
    """Cache instance's method during its life-time.
    Note: Order of the arguments is important. Different order of the arguments will result in different cache key.
    """

    def wrapper(func):
        fn_name = func.__name__

        @functools.wraps(func)
        def fn(self, *args, **kwargs):
            if not hasattr(self, "_cache"):
                self._cache = {}
            k = (
                fn_name,
                args,
                tuple(sorted(f"kw" + str(arg) for kw, arg in kwargs.items())),
            )
            if k not in self._cache:
                self._cache[k] = func(self, *args)
            return self._cache[k]

        return fn

    return wrapper



import glob
import os
import subprocess
from dataclasses import dataclass
from pathlib import Path
from typing import List, Optional, Union, NamedTuple
from pbt.misc import exec


class GitFileStatus(NamedTuple):
    is_deleted: bool
    fpath: str


class GitBranch(NamedTuple):
    local: Optional[str]
    remote: Optional[str]
    is_active: bool

    def get_name(self) -> str:
        if self.local is not None:
            return self.local.rsplit("/", 1)[-1]
        assert self.remote is not None
        return self.remote.rsplit("/", 1)[-1]

    def is_local(self) -> bool:
        return self.local is not None and self.local.find("/") == -1

    def get_remote(self) -> Optional[str]:
        if self.local is not None:
            if self.local.find("/") != -1:
                return self.local.split("/")[0]
            if self.remote is None:
                return None
        assert self.remote is not None
        return self.remote.rsplit("/")[1]


PathOrStr = Union[str, Path]


class Git:
    @classmethod
    def get_new_modified_deleted_files(cls, cwd: PathOrStr):
        (git_dir,) = exec("git rev-parse --show-toplevel", cwd=cwd)
        assert Path(git_dir).exists()

        output = subprocess.check_output(
            ["git", "status", "-uall", "--porcelain=v1", "--no-renames", "."], cwd=cwd
        ).decode()
        # rstrip as the first line can have empty character
        output = output.rstrip()
        if len(output) == 0:
            return []

        lines = output.split("\n")
        results = []

        # TODO: take a look at this document and implement it correctly
        #  https://git-scm.com/docs/git-status
        for line in lines:
            code = line[:2]
            rel_file_path = line[3:].strip()
            assert " -> " not in rel_file_path and "R" not in code
            results.append(
                GitFileStatus(
                    is_deleted="D" in code, fpath=os.path.join(git_dir, rel_file_path)
                )
            )
        return results

    @classmethod
    def get_current_commit(cls, cwd: PathOrStr):
        return (
            subprocess.check_output(["git", "rev-parse", "HEAD"], cwd=cwd)
            .decode()
            .strip()
        )

    @classmethod
    def get_current_branch(cls, cwd: PathOrStr):
        return (
            subprocess.check_output(
                ["git", "rev-parse", "--abbrev-ref", "HEAD"], cwd=cwd
            )
            .decode()
            .strip()
        )

    @classmethod
    def does_branch_exist(cls, cwd: PathOrStr, branch: str) -> bool:
        return (
            subprocess.check_output(["git", "branch", "--list", branch], cwd=cwd)
            .decode()
            .strip()
            == branch
        )

    @classmethod
    def checkout_branch(
        cls,
        cwd: PathOrStr,
        branch: str,
        create: bool = False,
        remote: Optional[str] = None,
    ):
        if not create:
            # exist local branch
            cmd = ["git", "checkout", branch]
        else:
            cmd = ["git", "checkout", "-b", branch]
            if remote is not None:
                cmd += ["--track", f"{remote}/{branch}"]
        subprocess.check_call(cmd, cwd=cwd)

    @classmethod
    def get_branches_contain_commit(cls, cwd: PathOrStr, commit_id: str):
        output = subprocess.check_output(
            ["git", "branch", "-a", "--contains", commit_id], cwd=cwd
        ).decode()
        branches = []

        for x in output.split("\n"):
            x = x.strip()
            if x == "":
                continue

            is_active = False
            if x[0] == "*":
                is_active = True
                x = x[2:]

            if x.startswith("(HEAD detached"):
                # detached head, so no branch
                continue

            if x.find(" -> ") != -1:
                remote, local = x.split(" -> ")
                assert (
                    remote.startswith("remotes/") and len(remote.split("/")) == 3
                ), f"Invalid remote: {remote}"
            elif x.startswith("remotes/") and len(x.split("/")) == 3:
                remote = x
                local = None
            else:
                remote = None
                local = x

            branch = GitBranch(local=local, remote=remote, is_active=is_active)
            if branch.get_name() == "HEAD":
                continue
            branches.append(branch)

        return branches

    @classmethod
    def auto_checkout_branch(cls, cwd: PathOrStr):
        commit_id = Git.get_current_commit(cwd)
        branches = Git.get_branches_contain_commit(cwd, commit_id)

        # select branch to checkout to, prefer development branch as we want to
        # resume previous work
        names = {branch.get_name() for branch in branches}
        if len(names.difference(["master", "main"])) > 0:
            name = list(names.difference(["master", "main"]))[0]
        else:
            name = "master" if "master" in names else "main"

        brs = [branch for branch in branches if branch.get_name() == name]
        has_local = any(br.is_local() for br in brs)
        if has_local:
            Git.checkout_branch(cwd, name)
        else:
            # try to identify the remote of this branch in case the branch is not tracked
            # prefer origin as it is the most common
            remotes = [r for r in {br.get_remote() for br in brs} if r is not None]
            if len(remotes) == 0:
                remote = None
            elif len(remotes) == 1:
                remote = remotes[0]
            elif "origin" in remotes:
                remote = "origin"
            else:
                raise Exception(
                    f"Cannot identify remote for branch {name} as there are multiple remotes: {remotes}"
                )

            if Git.does_branch_exist(cwd, name):
                # the local branch does exist, so we only need to checkout and pull
                Git.checkout_branch(cwd, name)
                Git.pull(cwd, remote=remote)
            else:
                Git.checkout_branch(cwd, name, create=True, remote=remote)

    @classmethod
    def init(cls, cwd: PathOrStr):
        subprocess.check_output(["git", "init"], cwd=cwd)

    @classmethod
    def push(cls, cwd: PathOrStr):
        exec("git push", cwd=cwd)

    @classmethod
    def pull(
        cls,
        cwd: PathOrStr,
        submodules: bool = False,
        remote: Optional[str] = None,
        verbose: bool = False,
    ):
        if verbose:
            fn = subprocess.check_call
        else:
            fn = subprocess.check_output

        try:
            fn(["git", "pull"], cwd=cwd)
        except:
            handle = False

            # if the the upstream branch does not set, we set it to the current remote
            try:
                r = subprocess.check_output(
                    [
                        "git",
                        "rev-parse",
                        "--abbrev-ref",
                        "--symbolic-full-name",
                        "@{u}",
                    ],
                    cwd=cwd,
                )
            except:
                branch = Git.get_current_branch(cwd)
                if remote is None:
                    # try to identify the remote
                    remotes = (
                        subprocess.check_output(["git", "remote", "-v"], cwd=cwd)
                        .decode()
                        .strip()
                        .split("\n")
                    )
                    if len(remotes) == 1:
                        remote = remotes[0]
                    elif "origin" in remotes:
                        remote = "origin"
                    else:
                        raise Exception(
                            f"Can't determine the correct remote for the branch as we have multiple remotes: {remotes}"
                        )

                subprocess.check_output(
                    [
                        "git",
                        "branch",
                        "--set-upstream-to",
                        f"{remote}/{branch}",
                        branch,
                    ],
                    cwd=cwd,
                )
                fn(["git", "pull"], cwd=cwd)
                handle = True

            if not handle:
                raise

        if submodules:
            fn(["git", "submodule", "update", "--init", "--recursive"], cwd=cwd)

    @classmethod
    def commit_all(cls, cwd: PathOrStr, msg: str = "add all files"):
        subprocess.check_output(["git", "add", "-A"], cwd=cwd)
        subprocess.check_output(["git", "commit", "-m", f"'{msg}'"], cwd=cwd)

    @classmethod
    def clone_all(cls, repo: str, cwd: PathOrStr) -> Path:
        """Clone a repository and its submodules. Then, return the cloned directory"""
        # get repo name to create directory to clone the repo into
        repo_name = repo.rsplit("/", 1)[-1].replace(".git", "")
        repo_dir = Path(cwd) / repo_name
        if repo_dir.exists():
            raise Exception(
                f"Directory {repo_dir} already exists. Can't clone the repository"
            )
        if not repo_dir.parent.exists():
            raise Exception(
                f"Directory {repo_dir.parent} does not exist. Can't clone the repository into it"
            )

        subprocess.check_call(
            [
                "git",
                "clone",
                "--recurse-submodules",
                "-j8",
                repo,
            ],
            cwd=cwd,
        )

        return repo_dir

    @classmethod
    def find_submodules(cls, repo_dir: PathOrStr) -> List[Path]:
        """Find submodules of a repository."""
        repo_dir = Path(os.path.abspath(repo_dir))

        submodules = [
            repo_dir / line.split(" ")[-1]
            for line in exec(
                "git config --file .gitmodules --get-regexp path", cwd=repo_dir
            )
        ]
        return submodules


if __name__ == "__main__":
    res = Git.get_new_modified_deleted_files("/workspace/sm-dev/osin")
    print("\n".join([str(x) for x in res]))



import importlib.metadata

import click, sys
from loguru import logger

from pbt.console.pbt import (
    obtain_prebuilt_binary,
    install,
    clean,
    install_local_pydep,
    update,
    publish,
    list_,
)
from pbt.console.git import git
from pbt.package.registry.pypi import PyPI


try:
    version = importlib.metadata.version("pbt")
except importlib.metadata.PackageNotFoundError:
    version = "0.0.0"


def check_upgrade():
    latest_version = PyPI.get_instance().get_latest_version("pbt")
    if latest_version is not None and version != latest_version:
        logger.warning(
            f"You are using an outdated version of pbt. The latest version is {latest_version}, while you are using {version}."
        )
    else:
        logger.trace("You are using the latest version of pbt.")


@click.group(
    help=f"PBT ({version}) -- a build tool for multi-projects that leverages package registries (pypi, npmjs, etc.)"
)
@click.version_option(version)
def cli():
    check_upgrade()


cli.add_command(install)
cli.add_command(clean)
cli.add_command(publish)
cli.add_command(update)
cli.add_command(git)
cli.add_command(list_)
cli.add_command(install_local_pydep)
cli.add_command(obtain_prebuilt_binary)

if __name__ == "__main__":
    cli()



import os
import shutil
from contextlib import contextmanager
from dataclasses import dataclass
from typing import Dict, Generator, List, Optional
from uuid import uuid4

import orjson
import semver
from loguru import logger
from hugedict.prelude import RocksDBDict, RocksDBOptions
from pbt.config import PBTConfig
from pbt.vcs.git import Git, GitFileStatus
from pbt.package.manager.manager import PkgManager
from pbt.package.package import Package

# file size limit
SOFT_SIZE_LIMIT = (1024**2) * 1  # 1MB
HARD_SIZE_LIMIT = (1024**2) * 100  # 100MBs
DIFF_DB_CACHE = {}


@contextmanager
def diff_db(
    pkg: Package, cfg: PBTConfig, new_connection: bool = False
) -> Generator["RocksDBDict[bytes, bytes]", None, None]:
    global DIFF_DB_CACHE
    db_file = str(cfg.pkg_cache_dir(pkg) / "diff.db")
    if new_connection:
        # client = DB(db_file, Options(create_if_missing=True))
        client = RocksDBDict(
            db_file,
            RocksDBOptions(create_if_missing=True),
            deser_key=bytes,
            deser_value=bytes,
            ser_value=lambda x: x,
        )
        try:
            yield client
        finally:
            del client
    else:
        if db_file not in DIFF_DB_CACHE:
            DIFF_DB_CACHE[db_file] = RocksDBDict(
                db_file,
                RocksDBOptions(create_if_missing=True),
                deser_key=bytes,
                deser_value=bytes,
                ser_value=lambda x: x,
            )
        yield DIFF_DB_CACHE[db_file]


def remove_diff_db(pkg: Package, cfg: PBTConfig):
    db_file = cfg.pkg_cache_dir(pkg) / "diff.db"
    if db_file.exists():
        shutil.rmtree(db_file)


@dataclass(eq=True)
class Diff:
    id: str
    commit_id: str
    changed_files: List[GitFileStatus]
    # content of the changed files, serving as a cache, does not guarantee to have
    # all of the non-deleted files, if the content of a changed file is None, it's
    # mean the value does not change since the last snapshot
    changed_files_content: Dict[str, Optional[bytes]]

    @staticmethod
    def from_local(
        db: "RocksDBDict[bytes, bytes]", manager: PkgManager, pkg: Package
    ) -> "Diff":
        """Compute diff of a current package, i.e., which files of a package have been modified"""
        commit_id = Git.get_current_commit(pkg.location)

        # TODO: make this code to get changed files more efficient
        changed_files = Git.get_new_modified_deleted_files(pkg.location)
        include_files = set(
            manager.filter_included_files(pkg, [file.fpath for file in changed_files])
        )
        changed_files = sorted(
            [file for file in changed_files if file.fpath in include_files]
        )

        return Diff(
            id=str(uuid4()),
            commit_id=commit_id,
            changed_files=changed_files,
            changed_files_content={},
        )

    def is_modified(self, db: "RocksDBDict[bytes, bytes]") -> bool:
        """Check if the package's content has been updated since the last snapshot"""
        prev_commit_id = db.get(b"commit_id")
        if prev_commit_id != self.commit_id.encode():
            return True

        prev_changed_files = orjson.loads(db[b"changed_files"])
        prev_changed_files = [GitFileStatus(*x) for x in prev_changed_files]
        if prev_changed_files != self.changed_files:
            return True

        for file in self.changed_files:
            if file.is_deleted:
                continue
            file_key = b"content:%s" % file.fpath.encode()
            prev_file_content = db.get(file_key)
            if file.fpath not in self.changed_files_content:
                file_content = read_file(file.fpath)
            else:
                file_content = self.changed_files_content[file.fpath]

            if prev_file_content != file_content:
                self.changed_files_content[file.fpath] = file_content
                return True
            self.changed_files_content[file.fpath] = None
        return False

    def save(self, db: "RocksDBDict[bytes, bytes]"):
        """Snapshot the current changes to the DB so that we can detect changes between commits"""
        db[b"commit_id"] = self.commit_id.encode()
        db[b"changed_files"] = orjson.dumps([tuple(x) for x in self.changed_files])

        prev_changed_files = db.get(b"changed_files")
        if prev_changed_files is not None:
            prev_changed_files = {
                GitFileStatus(*x).fpath for x in orjson.loads(prev_changed_files)
            }
            for file in prev_changed_files.difference(
                (x.fpath for x in self.changed_files)
            ):
                del db[b"content:%s" % file.encode()]

        for file in self.changed_files:
            if file.is_deleted:
                del db[b"content:%s" % file.fpath.encode()]
                continue

            if file.fpath not in self.changed_files_content:
                file_content = read_file(file.fpath)
            else:
                file_content = self.changed_files_content[file.fpath]
                if file_content is None:
                    # does not change the value since last snapshot, skip it
                    continue
            db[b"content:%s" % file.fpath.encode()] = file_content


@dataclass(eq=True)
class RemoteDiff:
    is_version_diff: bool
    is_content_changed: bool

    @staticmethod
    def from_pkg(
        manager: PkgManager,
        pkg: Package,
        remote_version: Optional[str] = None,
        remote_version_hash: Optional[str] = None,
    ) -> "RemoteDiff":
        """Detect if content of the current package is different from the remote package."""
        if pkg.version != remote_version:
            pkg_ver = semver.VersionInfo.parse(pkg.version)
            if remote_version is not None:
                remote_ver = semver.VersionInfo.parse(remote_version)
                if pkg_ver < remote_ver:
                    raise Exception(
                        "Current package version is outdated compared to the remote version"
                    )
            return RemoteDiff(is_version_diff=True, is_content_changed=True)

        # TODO: replace the approximation algorithm (checking hash) with exact algorithm
        # that determine if the content is the same
        pkg_hash = manager.compute_pkg_hash(pkg)
        return RemoteDiff(
            is_version_diff=False, is_content_changed=pkg_hash != remote_version_hash
        )


def read_file(fpath: str) -> bytes:
    fsize = os.path.getsize(fpath)
    if fsize > HARD_SIZE_LIMIT:
        raise Exception(
            f"File {fpath} is bigger than the hard limit ({format_size(fsize)} > {format_size(HARD_SIZE_LIMIT)})"
        )

    if fsize > SOFT_SIZE_LIMIT:
        logger.warning(
            "File {} is quite big ({} > {}). Consider ignore or commit it to speed up the process",
            fpath,
            format_size(fsize),
            format_size(SOFT_SIZE_LIMIT),
        )

    with open(fpath, "rb") as f:
        return f.read()


def format_size(n_bytes: int) -> str:
    if n_bytes < 1024:
        size = n_bytes
        unit = "Bs"
    elif n_bytes >= 1024 and n_bytes < (1024**2):
        size = round(n_bytes / 1024, 2)
        unit = "KBs"
    elif n_bytes >= (1024**2) and n_bytes < (1024**3):
        size = round(n_bytes / (1024**2), 2)
        unit = "MBs"
    else:
        assert n_bytes >= (1024**3)
        size = round(n_bytes / (1024**3), 2)
        unit = "GBs"
    return f"{size}{unit}"



from typing import Any, Optional, Dict, Tuple
import requests
import semver
from pbt.package.manager.manager import PkgManager

from pbt.package.registry.registry import PkgRegistry

PYPI_INDEX = "https://pypi.org"


class PyPI(PkgRegistry):
    instances: Dict[str, "PyPI"] = {}

    def __init__(self, index: str):
        self.index = index
        self.pkgs: Dict[str, Optional[dict]] = {}

    @staticmethod
    def get_instance(index: str = PYPI_INDEX) -> "PyPI":
        if index not in PyPI.instances:
            PyPI.instances[index] = PyPI(index)
        return PyPI.instances[index]

    def does_package_exist(
        self, pkg_name: str, package_version: Optional[str] = None
    ) -> bool:
        """Check if a package exist in pypi index"""
        pkg_info = self.fetch_pkg_info(pkg_name)
        if pkg_info is None:
            return False
        return package_version is None or package_version in pkg_info["releases"]

    def get_whl_hash(self, pkg_name: str, pkg_version: str) -> Optional[str]:
        """Get hash of a package (wheel distribution) at specific version"""
        pkg_info = self.fetch_pkg_info(pkg_name)
        if pkg_info is None:
            return None
        releases = pkg_info["releases"]
        if pkg_version not in releases:
            return None

        lst = [
            release
            for release in releases[pkg_version]
            if release["filename"].endswith(".whl")
        ]
        if len(lst) != 1:
            raise Exception(
                "Can't obtain hash of package %s as it does not have wheel release"
                % (pkg_name)
            )
        return lst[0]["digests"]["sha256"]

    def get_latest_version(self, pkg_name: str) -> Optional[str]:
        pkg_info = self.fetch_pkg_info(pkg_name)
        if pkg_info is None:
            return None

        releases: Dict[str, Any] = pkg_info["releases"]
        latest_version = max(releases.keys(), key=PkgManager.parse_version)
        return latest_version

    def get_latest_version_and_hash(self, pkg_name: str) -> Optional[Tuple[str, str]]:
        pkg_info = self.fetch_pkg_info(pkg_name)
        if pkg_info is None:
            return None

        releases: Dict[str, Any] = pkg_info["releases"]
        latest_version = max(releases.keys(), key=PkgManager.parse_version)
        whl_hash = self.get_whl_hash(pkg_name, latest_version)
        assert whl_hash is not None
        return latest_version, whl_hash

    def fetch_pkg_info(self, pkg_name: str) -> Optional[dict]:
        if pkg_name not in self.pkgs:
            resp = requests.get(self.index + f"/pypi/{pkg_name}/json")
            if resp.status_code == 404:
                self.pkgs[pkg_name] = None
            else:
                assert resp.status_code == 200
                self.pkgs[pkg_name] = resp.json()

        return self.pkgs[pkg_name]



import os
import zipfile
from pathlib import Path
import shutil
from typing import Dict, List, Literal, Tuple, cast

import click
from loguru import logger
import orjson

from pbt.config import PBTConfig
from pbt.package.manager.poetry import Poetry
from pbt.package.manager.maturin import Maturin
from pbt.package.manager.python import PythonPkgManager
from pbt.package.package import PackageType
from pbt.package.pipeline import BTPipeline, VersionConsistent
from pbt.package.registry.pypi import PyPI
from pbt.package.manager.manager import PkgManager, build_cache
from pbt.misc import exec


def init(
    cwd: str, packages: List[str], verbose: bool
) -> Tuple[BTPipeline, PBTConfig, List[str]]:
    cfg = PBTConfig.from_dir(cwd)
    managers: Dict[PackageType, PkgManager] = {
        PackageType.Poetry: Poetry(cfg),
        PackageType.Maturin: Maturin(cfg),
    }
    for k, v in managers.items():
        if isinstance(v, PythonPkgManager):
            v.set_package_managers(managers)

    pl = BTPipeline(cfg, managers=managers)
    pl.discover()

    if len(packages) == 0:
        pkgs = set(pl.pkgs.keys())
    else:
        pkgs = set(packages)

    if len(pkgs.difference(pl.pkgs.keys())) > 0:
        raise Exception(
            f"Passing unknown packages: {pkgs.difference(pl.pkgs.keys())}. Available options: {list(pl.pkgs.keys())}"
        )

    return pl, cfg, sorted(pkgs)


@click.command(name="list")
@click.option(
    "-d",
    "--dev",
    is_flag=True,
    help="Whether to print to the local (inter-) dependencies",
)
@click.option("--cwd", default=".", help="Override current working directory")
@click.option(
    "-v",
    "--verbose",
    is_flag=True,
    help="increase verbosity",
)
def list_(dev: bool = False, cwd: str = ".", verbose: bool = False):
    """List all packages in the current project, and their dependencies if required."""
    pl, cfg, pkgs = init(cwd, [], verbose)
    for pkg in pl.pkgs.values():
        print_children = (
            dev and any(True for dep in pkg.dependencies.keys() if dep in pl.pkgs) > 0
        )
        print(f"{pkg.name} ({pkg.version})" + (":" if print_children else ""))
        if dev:
            for dep in pkg.dependencies.keys():
                if dep in pkgs:
                    print(f"\t- {dep} ({pl.pkgs[dep].version})")


@click.command()
@click.option(
    "-p",
    "--package",
    multiple=True,
    help="Specify the package that we want to build. If empty, build all packages",
)
@click.option(
    "-d",
    "--dev",
    is_flag=True,
    help="Whether to install dev-dependencies as well",
)
@click.option("--cwd", default=".", help="Override current working directory")
@click.option(
    "-v",
    "--verbose",
    is_flag=True,
    help="increase verbosity",
)
def install(
    package: List[str],
    dev: bool = False,
    cwd: str = ".",
    verbose: bool = False,
):
    """Make package"""
    pl, cfg, pkgs = init(cwd, package, verbose)
    pl.enforce_version_consistency(freeze_packages=cfg.freeze_packages)
    pl.install(sorted(pkgs), include_dev=dev)


@click.command()
@click.option(
    "-p",
    "--package",
    multiple=True,
    help="Specify the package that we want to build. If empty, build all packages",
)
@click.option("--cwd", default=".", help="Override current working directory")
@click.option(
    "-v",
    "--verbose",
    is_flag=True,
    help="increase verbosity",
)
def clean(package: List[str], cwd: str = ".", verbose: bool = False):
    """Clean packages' build & lock files"""
    pl, cfg, pkgs = init(cwd, package, verbose)
    pl.enforce_version_consistency(freeze_packages=cfg.freeze_packages)
    for pkg_name in pkgs:
        pkg = pl.pkgs[pkg_name]
        pl.managers[pkg.type].clean(pkg)


@click.command()
@click.option("--cwd", default=".", help="Override current working directory")
@click.option(
    "-v",
    "--verbose",
    is_flag=True,
    help="increase verbosity",
)
def update(cwd: str = ".", verbose: bool = False):
    """Update all package inter-dependencies"""
    pl, cfg, pkgs = init(cwd, [], verbose)
    pl.enforce_version_consistency(
        VersionConsistent.STRICT, freeze_packages=cfg.freeze_packages
    )


@click.command()
@click.option(
    "-p",
    "--package",
    multiple=True,
    help="Specify the package that we want to build. If empty, build all packages",
)
@click.option("--cwd", default=".", help="Override current working directory")
@click.option(
    "-v",
    "--verbose",
    is_flag=True,
    help="increase verbosity",
)
def publish(package: List[str], cwd: str = ".", verbose: bool = False):
    """Publish packages"""
    pl, cfg, pkgs = init(cwd, package, verbose)
    pl.enforce_version_consistency(freeze_packages=cfg.freeze_packages)
    pl.publish(
        pkgs,
        {
            PackageType.Poetry: PyPI.get_instance(),
        },
    )


@click.command()
@click.option(
    "-p",
    "--package",
    multiple=True,
    help="Specify the package that we want to build. If empty, build all packages",
)
@click.option("--cwd", default=".", help="Override current working directory")
@click.option(
    "-v",
    "--verbose",
    is_flag=True,
    help="increase verbosity",
)
def build(package: List[str], cwd: str = ".", verbose: bool = False):
    """Build packages"""
    pl, cfg, pkgs = init(cwd, package, verbose)
    with build_cache():
        for pkg_name in pkgs:
            pkg = pl.pkgs[pkg_name]
            manager = pl.managers[pkg.type]
            manager.build(pkg)


@click.command()
@click.option(
    "-p",
    "--package",
    help="The parent package",
)
@click.option(
    "-d",
    "--dep",
    help="The dependency package",
)
@click.option(
    "--include-dep-deps",
    is_flag=True,
    help="Whether to also install the dependencies of the dependency",
)
@click.option("--cwd", default=".", help="Override current working directory")
@click.option(
    "-v",
    "--verbose",
    is_flag=True,
    help="increase verbosity",
)
def install_local_pydep(
    package: str,
    dep: str,
    include_dep_deps: bool,
    cwd: str = ".",
    verbose: bool = False,
):
    """Install a local python package in editable mode without building it. This works by adding .pth file containing the path
    to the package to the site-packages directory.

    This is a temporary solution for package requiring extension binary but we cannot build the binary,
    so we have to download the prebuilt binary and put it to the src directory before running this command.

    Note: this command won't enforce version consistency as it is intended to be used in hacky situations.
    """
    pl, cfg, pkgs = init(cwd, [package], verbose)
    pkg = pl.pkgs[package]
    dep_pkg = pl.pkgs[dep]

    logger.info(
        "Installing {} in editable mode inside {}'s environment", dep_pkg.name, pkg.name
    )
    if include_dep_deps:
        logger.info("Also installing {}'s dependencies", dep_pkg.name)
        skip_dep_deps = [
            name for name in dep_pkg.get_all_dependency_names() if name in pl.pkgs
        ]
    else:
        skip_dep_deps = dep_pkg.get_all_dependency_names()

    manager = pl.managers[pkg.type]
    assert isinstance(manager, PythonPkgManager) and isinstance(
        pl.managers[dep_pkg.type], PythonPkgManager
    )

    if dep_pkg.type == PackageType.Maturin:
        # manually create a pyproject.toml in poetry first
        try:
            os.rename(
                dep_pkg.location / "pyproject.toml",
                dep_pkg.location / "pyproject.origin.toml",
            )
            pl.managers[PackageType.Poetry].save(dep_pkg)
            dep_pkg.type = PackageType.Poetry
            manager.install_dependency(
                pkg,
                dep_pkg,
                skip_dep_deps=skip_dep_deps,
            )
        finally:
            os.rename(
                dep_pkg.location / "pyproject.origin.toml",
                dep_pkg.location / "pyproject.toml",
            )
    else:
        manager.install_dependency(
            pkg,
            dep_pkg,
            skip_dep_deps=skip_dep_deps,
        )


@click.command()
@click.option(
    "-p",
    "--package",
    required=True,
    help="The package to extract the binary from",
)
@click.option("--cwd", default=".", help="Override current working directory")
@click.option(
    "-v",
    "--verbose",
    is_flag=True,
    help="increase verbosity",
)
@click.option(
    "-s",
    "--source",
    type=click.Choice(["build", "pypi", "build-dev"], case_sensitive=False),
    default="build",
    help="source where we can pull/build the dependency",
)
@click.option(
    "-c",
    "--clean",
    is_flag=True,
    help="whether to clean the dist directory after this command",
)
def obtain_prebuilt_binary(
    package: str,
    cwd: str = ".",
    verbose: bool = False,
    source: Literal["build", "pypi"] = "build",
    clean: bool = False,
):
    pl, cfg, pkgs = init(cwd, [package], verbose)
    pkg = pl.pkgs[package]

    manager = pl.managers[pkg.type]
    assert isinstance(manager, PythonPkgManager)

    dist_dir = (pkg.location / cfg.distribution_dir).absolute()
    shutil.rmtree(dist_dir, ignore_errors=True)
    dist_dir.mkdir()

    if source == "build" or source == "build-dev":
        manager._build_command(pkg, release=source == "build")
    else:
        exec(
            ["pip", "download", pkg.name, "--no-deps"],
            cwd=dist_dir,
            env=manager.passthrough_envs,
            redirect_stderr=True,
        )

    (whl_file,) = [x for x in dist_dir.glob("*.whl")]
    with zipfile.ZipFile(whl_file, "r") as zip_ref:
        zip_ref.extractall(dist_dir)

    pkg_name = pkg.name.replace("-", "_")
    pkg_dir = dist_dir / pkg_name
    if not pkg_dir.exists():
        for name in pkg.include:
            pkg_name = name.replace("-", "_")
            pkg_dir = dist_dir / pkg_name
            if pkg_dir.exists():
                break
        else:
            raise RuntimeError(f"Cannot find the package directory in {dist_dir}")

    for file in pkg_dir.iterdir():
        if file.name.endswith(".so") or file.name.endswith(".dylib"):
            dest_file = pkg.location / pkg_name / file.name
            if dest_file.exists():
                dest_file.unlink()
            shutil.move(file, pkg.location / pkg_name)

    if clean:
        shutil.rmtree(dist_dir)



import os
from dataclasses import dataclass, field
from pathlib import Path
from typing import List, Optional, Set, Union
from loguru import logger

import orjson
from pbt.package.package import Package
from pbt.misc import cache_method, exec


PBT_CONFIG_FILE_NAME = "pbtconfig.json"
PBT_LOCK_FILE_NAME = "pbt.lock"
PBT_IGNORE_FILE_NAME = ".pbtignore"


@dataclass
class PBTConfig:
    # current working directory
    cwd: Path
    # directory
    cache_dir: Path
    # set of packages that we ignore
    ignore_packages: Set[str] = field(default_factory=set)
    # set of directories (absolute path) that we ignore and not search for packages in them
    ignore_directories: Set[Path] = field(default_factory=set)
    # set of directory name that we ignore and not search for packages in them
    ignore_directory_names: Set[str] = field(default_factory=set)
    # packages that do not contain any code with sole purpose for installing dependencies or creating working environment
    phantom_packages: Set[str] = field(default_factory=set)
    # use pre-built binaries for the package if available (i.e., rely on the package registry to find an installable version)
    use_prebuilt_binaries: Set[str] = field(default_factory=set)
    # freeze the following packages and don't update their dependencies
    freeze_packages: Set[str] = field(default_factory=set)
    # directory to store the built artifacts for release (relative to each package's location)
    distribution_dir: Path = Path("./dist")
    # the virtualenv directory (default is .venv in the project root directory)
    python_virtualenvs_path: str = "./.venv"
    # python executable to use for building and installing packages, default (None) is the first one on PATH
    python_path: Optional[Path] = None

    @staticmethod
    def from_dir(cwd: Union[Path, str]) -> "PBTConfig":
        # get git top module
        # TODO: replace me
        try:
            output = exec(
                "git rev-parse --show-superproject-working-tree --show-toplevel",
                cwd="." if cwd == "" else cwd,
            )
            if len(output) == 1:
                cwd = output[0]
            else:
                for i in range(len(output)):
                    if all(
                        output[j].startswith(output[i])
                        for j in range(i + 1, len(output))
                    ):
                        cwd = output[i]
                        break
                else:
                    raise Exception(
                        "Unreachable error. Can't figure out which folder contains your project. Congrat! You found a bug.\nAvailable options:\n"
                        + "\n\t".join(output)
                    )
        except Exception as e:
            if not str(e).startswith("fatal: not a git repository"):
                # another error not related to git
                raise
            else:
                cwd = cwd

        cwd = Path(cwd).absolute()
        cache_dir = cwd / ".cache"
        cache_dir.mkdir(exist_ok=True, parents=True)

        if (cwd / PBT_CONFIG_FILE_NAME).exists():
            with open(cwd / PBT_CONFIG_FILE_NAME, "r") as f:
                cfg = orjson.loads(f.read())
        else:
            cfg = {}

        if (cwd / PBT_IGNORE_FILE_NAME).exists():
            with open(cwd / PBT_IGNORE_FILE_NAME, "r") as f:
                instructions = {x.strip() for x in f.readlines()}
                if "" in instructions:
                    instructions.remove("")
                ignore_directories = set()
                ignore_directory_names = set()
                for instruction in instructions:
                    if instruction.startswith("/"):
                        ignore_directories.add((cwd / instruction[1:]).resolve())
                    else:
                        assert "/" not in instruction, instruction
                        ignore_directory_names.add(instruction)
        else:
            ignore_directories = set()
            ignore_directory_names = set()

        logger.info("Root directory: {}", cwd)

        if "python_path" not in cfg:
            # try use python_path from the environment variable: `PBT_PYTHON`
            python_path = os.environ.get("PBT_PYTHON", None)
        else:
            python_path = cfg["python_path"]

        return PBTConfig(
            cwd=cwd,
            cache_dir=cache_dir,
            ignore_packages=set(cfg.get("ignore_packages", [])),
            ignore_directories=ignore_directories,
            ignore_directory_names=ignore_directory_names,
            phantom_packages=set(cfg.get("phantom_packages", [])),
            use_prebuilt_binaries=set(cfg.get("use_prebuilt_binaries", [])),
            freeze_packages=set(cfg.get("freeze_packages", [])),
            distribution_dir=Path(cfg.get("distribution_dir", "./dist")),
            python_virtualenvs_path=cfg.get("python_virtualenvs_path", "./.venv"),
            python_path=Path(python_path) if python_path is not None else None,
        )

    def pkg_cache_dir(self, pkg: Package) -> Path:
        """Get the cache directory for a package that we can use for storing temporary files
        during building and installing packages
        """
        pkg_dir = self.cache_dir / pkg.name
        pkg_dir.mkdir(exist_ok=True, parents=True)
        return pkg_dir

    @cache_method()
    def get_python_path(self) -> str:
        if self.python_path is not None:
            if not self.python_path.exists():
                raise ValueError("Python does not exist: {}".format(self.python_path))
            return str(self.python_path)
        else:
            return "python"



from __future__ import annotations

from dataclasses import dataclass
from itertools import chain
import networkx as nx
from typing import Dict, Iterable, List, Optional, Type, Union

from pbt.package.package import Package, DepConstraints, PackageType


@dataclass
class ThirdPartyPackage:
    name: str
    type: PackageType
    # mapping from source package that use this package to the version the source package depends on
    invert_dependencies: Dict[str, DepConstraints]


class PkgGraph:
    """Representing the dependencies between packages (including third-party packages) in the project.
    The edge between (A, B) represents a dependency relationship that package A uses package B.
    """

    def __init__(self, g: Optional[nx.DiGraph] = None) -> None:
        self.g = g or nx.DiGraph()
        self.pkgs = {}

    @staticmethod
    def from_pkgs(pkgs: Dict[str, Package]) -> PkgGraph:
        """Create a PkgGraph from a dictionary of packages.

        Args:
            pkgs: A dictionary of packages.
        """
        g = nx.DiGraph()
        for pkg in pkgs.values():
            g.add_node(pkg.name, pkg=pkg)

        for pkg in pkgs.values():
            for deps, is_dev in [
                (pkg.dependencies, False),
                (pkg.dev_dependencies, True),
            ]:
                deps: Dict[str, DepConstraints]
                for dep, specs in deps.items():
                    if not g.has_node(dep):
                        assert dep not in pkgs
                        g.add_node(
                            dep,
                            pkg=ThirdPartyPackage(dep, pkg.type, {pkg.name: specs}),
                        )
                    else:
                        dep_pkg = g.nodes[dep]["pkg"]
                        if isinstance(dep_pkg, ThirdPartyPackage):
                            assert dep_pkg.type.is_compatible(pkg.type)
                            dep_pkg.invert_dependencies[pkg.name] = specs
                    g.add_edge(pkg.name, dep, is_dev=is_dev)

        try:
            cycles = nx.find_cycle(g, orientation="original")
            raise ValueError(f"Found cyclic dependencies: {cycles}")
        except nx.NetworkXNoCycle:
            pass

        return PkgGraph(g)

    def iter_pkg(self) -> Iterable[Union[ThirdPartyPackage, Package]]:
        """Iterate over all packages in the graph."""
        return (self.g.nodes[pkg]["pkg"] for pkg in self.g.nodes)

    def dependencies(
        self, pkg_name: str, include_dev: bool = False
    ) -> List[Union[ThirdPartyPackage, Package]]:
        """Return the list of packages that are dependency of the input package.

        Args:
            pkg_name: The package to get the dependencies of.
            include_dev: Whether to include dev dependencies.
        """
        nodes = list(nx.dfs_preorder_nodes(self.g, pkg_name))
        assert nodes[0] == pkg_name, "The first node is the package itself"
        nodes = nodes[1:]

        if include_dev:
            return [self.g.nodes[uid]["pkg"] for uid in nodes]

        lst = []
        for vid in nodes:
            predecessors = self.g.predecessors(vid)
            if all(self.g[uid][vid]["is_dev"] for uid in predecessors):
                continue
            lst.append(self.g.nodes[vid]["pkg"])
        return lst



import glob
import os
import re
from abc import ABC, abstractmethod
from contextlib import contextmanager
from pathlib import Path
from typing import Dict, List, Literal, Optional, Sequence, Set, Tuple, Union

import semver
from pbt.config import PBTConfig
from pbt.misc import cache_func, cache_method
from pbt.package.package import DepConstraint, DepConstraints, Package, VersionSpec


class PkgManager(ABC):
    """A package manager that is responsible for all tasks related to packages such as dicovering, parsing, building, installing, and publishing."""

    def __init__(self, cfg: PBTConfig) -> None:
        self.cfg = cfg

    @abstractmethod
    def is_package_directory(self, dir: Path) -> bool:
        """Check if the given directory is a package directory."""
        raise NotImplementedError()

    @abstractmethod
    def glob_query(self, root: Path) -> str:
        """Return glob query to iterate over all (potential) packages in the project

        Example:
            - "{root}/**/package.json" for NodeJs
            - "{root}/**/package.xml" for Maven
            - "{root}/**/pyproject.toml" for Poetry

        Args:
            root_dir: The root directory where we start searching

        Returns:
            A glob query to iterate over all specification files in the project
        """
        raise NotImplementedError()

    def discover(
        self, root: Path, ignore_dirs: Set[Path], ignore_dirnames: Set[str]
    ) -> List[Path]:
        """Discover potential packages in the project

        Args:
            root: The root directory to start searching
            ignore_dirs: Do not search in these directories
        """
        out = [Path(f) for f in glob.glob(self.glob_query(root.resolve()))]
        if len(ignore_dirs) > 0:
            out = [
                f for f in out if not f in ignore_dirs and f.name not in ignore_dirnames
            ]
        return out

    @abstractmethod
    def load(self, dir: Path) -> Package:
        """Load a package from the given directory.

        Args:
            dir: The directory where the package is located
        """
        raise NotImplementedError()

    @abstractmethod
    def save(self, package: Package):
        """Save the current configuration of the package

        Args:
            package: The package to save
        """
        raise NotImplementedError()

    @abstractmethod
    def clean(self, package: Package):
        """Remove previously installed dependencies and the environment where the package is installed, for a freshly start.
        In addition, this also cleans built artifacts.

        Args:
            package: The package to clean
        """
        raise NotImplementedError()

    @abstractmethod
    def publish(self, package: Package):
        """Publish the package to the package registry

        Args:
            package: The package to publish
        """
        raise NotImplementedError()

    @abstractmethod
    def build(
        self,
        package: Package,
        skip_deps: Optional[List[str]] = None,
        additional_deps: Optional[Dict[str, DepConstraints]] = None,
        release: bool = True,
        clean_dist: bool = True,
    ):
        """Build the package. If it has been built before during PBT running, we may skip the build step (i.e., caching results).

        Args:
            package: The package to build
            skip_deps: The optional list of dependencies to skip
            additional_deps: additional dependencies to add to the build
            release: whether to build in release mode
            clean_dist: Whether to clean the dist directory before building
        """
        raise NotImplementedError()

    @abstractmethod
    def install_dependency(
        self,
        package: Package,
        dependency: Package,
        skip_dep_deps: Optional[List[str]] = None,
    ):
        """Install the given dependency for the given package.

        The dependency must not be the phantom package as it does not containing any code to build or install.

        Note: don't expect this function will be able to find the local dependencies in the project
        as the manager relies on a package registry. For local dependencies, add them to `skip_deps` and use `install_dependency` to install them
        separately instead. Otherwise, you may get an error or the local dependencies from the package registry (the content may be different from the current ones on disk).

        Args:
            package: The package to install the dependency for
            dependency: The dependency to install
            skip_dep_deps: The dependencies of the dependency to skip. This option is not guaranteed to compiled language
        """
        raise NotImplementedError()

    @abstractmethod
    def install(
        self,
        package: Package,
        include_dev: bool = False,
        skip_deps: Optional[List[str]] = None,
        additional_deps: Optional[Dict[str, DepConstraints]] = None,
    ):
        """Install the package, assuming the the specification is updated. Note that if the package is phantom, only the dependencies are installed.

        Note: don't expect this function will be able to find the local dependencies in the project
        as the manager relies on a package registry. For local dependencies, add them to `skip_deps` and use `install_dependency` to install them
        separately instead. Otherwise, you may get an error.

        Args:
            package: The package to install
            include_dev: Whether to install dev dependencies
            skip_deps: The dependencies to skip (usually the local ones we want to install in editable mode separately). This option is not guaranteed to compiled language
            additional_deps: The additional dependencies to install
        """
        raise NotImplementedError()

    @abstractmethod
    def compute_pkg_hash(self, pkg: Package, target: Optional[str] = None) -> str:
        """Compute hash of package's content.

        Args:
            pkg: The package to compute the hash for
            target: (optional) a specific target that this package is built for. The format and value depends on the kind of package.
                For example, in Python it is: `(-{build tag})?-{python tag}-{abi tag}-{platform tag}.whl` (pep-491)
        """
        raise NotImplementedError()

    @abstractmethod
    def get_fixed_version_pkgs(self) -> Set[str]:
        """Get set of packages which versions are fixed and never should updated."""
        raise NotImplementedError()

    @contextmanager  # type: ignore
    def mask_file(
        self,
        file_path: Union[str, Path],
    ):
        """Temporary mask out a file

        Arguments:
            file_path: The path to the file to mask
        """
        file_path = str(file_path)
        assert os.path.isfile(file_path)
        try:
            os.rename(file_path, file_path + ".tmp")
            yield None
        finally:
            os.rename(file_path + ".tmp", file_path)

    def filter_included_files(self, pkg: Package, files: List[str]) -> List[str]:
        """Filter out the files that are not included in the package, these are files that
        won't affect the content of the package.

        Args:
            pkg: The package to check against
            files: The files to filter
        """
        dir_paths = []
        patterns = []

        for pattern in pkg.include:
            pattern = os.path.join(pkg.location, pattern)
            # TODO: detect the pattern better
            if "*" in pattern or "?" in pattern or ("[" in pattern and "]" in pattern):
                patterns.append(pattern)
            else:
                dir_paths.append(pattern)

        # include nested packages
        for depfile in glob.glob(self.glob_query(pkg.location)):
            dir_paths.append(depfile)

        if len(patterns) > 0:
            for pattern in patterns:
                if not (pattern.endswith("/**/*") and pattern[:-5].find("*") == -1):
                    raise NotImplementedError()
                dir_paths.append(pattern[:-5])

        output = []
        for file in files:
            if any(file.startswith(dpath) for dpath in dir_paths):
                output.append(file)
        return output

    def find_latest_specs(
        self,
        lst_specs: List[DepConstraints],
        mode: Optional[Literal["strict", "compatible"]] = None,
    ) -> DepConstraints:
        """Given a set of specs, some of them may be the same, some of them are older.

        This function finds the latest version for each constraint based on the lowerbound

        Args:
            lst_specs: list of list of specs
            mode: 'strict' or 'compatible' or None.
                'strict' means all versions must be the same.
                'compatible' means all versions must be compatible.
                None means we don't check
        """
        constraints: Dict[Optional[str], Tuple[VersionSpec, DepConstraint]] = {}
        for specs in lst_specs:
            for spec in specs:
                version_spec = self.parse_version_spec(spec.version_spec)
                if spec.constraint not in constraints:
                    constraints[spec.constraint] = (version_spec, spec)
                else:
                    prev_version_spec, prev_spec = constraints[spec.constraint]
                    if version_spec == prev_version_spec:
                        continue

                    has_recent_version = False

                    # determine if the current spec is newer than the previous one
                    if (
                        version_spec.lowerbound is None
                        or prev_version_spec.lowerbound is None
                    ):
                        raise ValueError(
                            f"Uncompatible constraint {spec.version_spec} vs {prev_spec}. Consider fixing it"
                        )

                    if mode == "strict":
                        raise ValueError()
                    elif mode == "compatible":
                        # they are incompatible if either the same or one is stricter
                        try:
                            version_spec.intersect(prev_version_spec)
                        except ValueError:
                            # they are incompatible
                            raise

                    if version_spec.lowerbound > prev_version_spec.lowerbound:
                        has_recent_version = True
                    elif version_spec.upperbound is not None and (
                        prev_version_spec.upperbound is None
                        or version_spec.upperbound < prev_version_spec.upperbound
                    ):
                        has_recent_version = True

                    if has_recent_version:
                        constraints[spec.constraint] = (version_spec, spec)

        # sort to make the order consistent
        return [v[1] for k, v in sorted(constraints.items(), key=lambda x: x[0] or "")]

    def is_version_compatible(
        self, version: semver.VersionInfo, version_spec: str
    ) -> bool:
        """Check if the given version is compatible with the given rule

        Args:
            version: package version
            version_spec: The version spec to check against
        """
        return self.parse_version_spec(version_spec).is_version_compatible(version)

    @classmethod
    @cache_func()
    def parse_version_spec(
        cls,
        rule: str,
    ) -> VersionSpec:
        """Parse the given version rule to get lowerbound and upperbound (exclusive)

        Example:
            - "^1.0.0" -> (1.0.0, 2.0.0)
            - ">= 1.0.0" -> (1.0.0, None)
            - ">= 1.0.0, < 2.1.3" -> (1.0.0, 2.1.3)
        """
        m = re.match(
            r"(?P<op1>\^|~|>|>=|==|<|<=)? *(?P<version1>[^ ,\^\~>=<]+)(?:(?:(?: *, *)|(?: +))(?P<op2>\^|~|>|>=|==|<|<=) *(?P<version2>[^ ,\^\~>=<]+))?",
            rule,
        )
        assert (
            m is not None
        ), f"The constraint is too complicated to handle for now: `{rule}`"

        op1, version1 = m.group("op1"), m.group("version1")
        op2, version2 = m.group("op2"), m.group("version2")

        if op1 == "":
            op1 = "=="

        lowerbound = cls.parse_version(version1)
        if op1 == "^":
            assert version2 is None
            # special case for 0 following the nodejs way (I can't believe why)
            # see more: https://nodesource.com/blog/semver-tilde-and-caret/
            if lowerbound.major == 0:
                if lowerbound.minor == 0:
                    upperbound = lowerbound.bump_patch()
                else:
                    upperbound = lowerbound.bump_minor()
            else:
                upperbound = lowerbound.bump_major()
            spec = VersionSpec(
                lowerbound=lowerbound,
                upperbound=upperbound,
                is_lowerbound_inclusive=True,
                is_upperbound_inclusive=False,
            )
        elif op1 == "~":
            assert version2 is None
            if m.group("version1").isdigit():
                # only contains major version
                upperbound = lowerbound.bump_major()
            else:
                upperbound = lowerbound.bump_minor()
            spec = VersionSpec(
                lowerbound=lowerbound,
                upperbound=upperbound,
                is_lowerbound_inclusive=True,
                is_upperbound_inclusive=False,
            )
        elif op1 == "==":
            assert version2 is None
            upperbound = lowerbound
            spec = VersionSpec(
                lowerbound=lowerbound,
                upperbound=upperbound,
                is_lowerbound_inclusive=True,
                is_upperbound_inclusive=True,
            )
        else:
            upperbound = cls.parse_version(version2) if version2 is not None else None
            if op1 == "<" or op1 == "<=":
                op1, op2 = op2, op1
                lowerbound, upperbound = upperbound, lowerbound
            spec = VersionSpec(
                lowerbound=lowerbound,
                upperbound=upperbound,
                is_lowerbound_inclusive=op1 == ">=",
                is_upperbound_inclusive=op2 == "<=",
            )
        return spec

    def update_version_spec(
        self, version_spec: str, version: Union[str, semver.VersionInfo]
    ) -> str:
        """Update the given version spec to compatible to the given version

        Args:
            version_spec: The version spec to update
            version: The version to update the version spec with
        """
        m = re.match(r"(\^|~|>|>=|==|<|<=) *([^ ]+)", version_spec)
        if m is None:
            complex_spec = True
            if version_spec[0].isdigit():
                try:
                    self.parse_version(version_spec)
                    complex_spec = False
                except AssertionError:
                    pass

            if complex_spec:
                raise NotImplementedError(
                    f"Not implementing update complicated version spec `{version_spec}` yet"
                )
            formula = "=="
        else:
            formula = m.groups()[0]
        return f"{formula}{str(version)}"

    @classmethod
    @cache_func()
    def parse_version(cls, version: str) -> semver.VersionInfo:
        m = re.match(
            r"^(?P<major>\d+)(?P<minor>\.\d+)?(?P<patch>\.\d+)?(?P<rest>[^\d].*)?$",
            version,
        )
        assert (
            m is not None
        ), f"Current parser is not able to parse version: `{version}` yet"

        if m is not None:
            parts = [
                m.group("major"),
                m.group("minor") or ".0",
                m.group("patch") or ".0",
                m.group("rest") or "",
            ]
            if not parts[-1].startswith("-") and parts[-1] != "":
                # add hyphen to make it compatible with semver package.
                # e.g. 21.11b1 -> 21.11.0-b1
                parts[-1] = "-" + parts[-1]
            version = "".join(parts)

        return semver.VersionInfo.parse(version)

    def next_version(self, pkg: Package, rule: Literal["major", "minor", "patch"]):
        """Update version of the package if a valid bump rule is provided

        Args:
            pkg: The package to update
            rule: The rule to update the version with
        """
        pkg.version = str(self.parse_version(pkg.version).next_version(rule))
        self.save(pkg)


@contextmanager
def build_cache(_cache={}):
    """Yield a dictionary that can be used to cache result of the build (whether the package has been built or not)

    Nested call will return the same dictionary.
    """
    if "db" not in _cache:
        _cache["db"] = {}
    if "count" not in _cache:
        _cache["count"] = 0

    try:
        _cache["count"] += 1
        yield _cache["db"]
    finally:
        _cache["count"] -= 1
        if _cache["count"] == 0:
            _cache["db"] = {}



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pbt',
 'pbt.console',
 'pbt.package',
 'pbt.package.manager',
 'pbt.package.registry',
 'pbt.vcs']

package_data = \
{'': ['*']}

install_requires = \
['click>=8.1.3,<9.0.0',
 'hugedict>=2.7.1,<3.0.0',
 'loguru>=0.6.0,<0.7.0',
 'maturin>=0.13.6,<0.14.0',
 'networkx>=2.8.2,<3.0.0',
 'orjson>=3.8.0,<4.0.0',
 'poetry>=1.2.2,<2.0.0',
 'requests>=2.28.0,<3.0.0',
 'semver>=2.13.0,<3.0.0',
 'tomlkit>=0.11.5,<0.12.0',
 'typing-extensions>=4.4.0,<5.0.0']

entry_points = \
{'console_scripts': ['pbt = pbt.__main__:cli']}

setup_kwargs = {
    'name': 'pbt',
    'version': '2.10.8',
    'description': 'A build tool for multiple Python projects in a single repository',
    'long_description': '<h1 align="center">PBT</h1>\n\n<div align="center">\n<b>pbt</b> — A build tool for multiple Python projects in a single repository\n    \n![PyPI](https://img.shields.io/pypi/v/pbt)\n![Python](https://img.shields.io/badge/python-v3.8+-blue.svg)\n[![GitHub Issues](https://img.shields.io/github/issues/binh-vu/pbt.svg)](https://github.com/binh-vu/pbt/issues)\n![Contributions welcome](https://img.shields.io/badge/contributions-welcome-orange.svg)\n[![License](https://img.shields.io/badge/license-MIT-blue.svg)](https://opensource.org/licenses/MIT)\n\n</div>\n\n## Introduction\n\nHaving all packages in the same repository make it much easier to develop, share, reuse, and refactor code. Building and publishing the packages should not be done manually because it is time-consuming and may be frustrated if the projects are depending on each other. [pbt](https://github.com/binh-vu/pbt) is a tool designed to help make the process easier and faster. It supports building, installing, and updating versions of your packages and their dependencies consistently. It also provides utility commands to help you work with your packages in multi-repositories as if you are working with a monorepo.\n\n## Installation\n\n```bash\npip install -U pbt\n```\n\n## Usage\n\nNote: currently, [pbt](https://github.com/binh-vu/pbt) supports Python packages configured with Poetry (an awesome dependency management that you should consider using).\n\nAssuming that you organized your packages to different sub-folders, each has their own project configuration file (e.g., `pyproject.toml`). You can run the following commands in the root directory (containing your projects). Note: [pbt](https://github.com/binh-vu/pbt) will discover the project based on the project name in its configuration file not the folder name.\n\nYou can also discover the list of commands by running `pbt --help`. Many commands have an option `--cwd` to override the current working directory.\n\n1. **List all packages in the current project, and their dependencies if required**\n\n```bash\npbt list [-d]\n```\n\n- `-d`, `--dev`: Whether to print to the local (inter-) dependencies\n\n2. **Create virtual environment of a package and install its dependencies**\n\n```bash\npbt install [-d] [-v] [-p <package>]\n```\n\n- `-d`: also install dev-dependencies of the package\n- `-v`: verbose\n- `-p`: specify the package we want to build, if empty build all packages.\n\nIf you have encounter some errors during the installation, you can checkout the `pyproject.failed.toml` file that is generated by pbt in `./cache/<package>` folder (relative to your current working directory). For example, on M1 chip, if your python version is `^3.8`, you can\'t use the newer scipy (e.g., >1.8 as it requires python `<3.11`), poetry lock chooses to use an old version `1.6.0`, which typically can\'t build on M1 due to no pre-built numpy for it.\n\n3. **Update all package inter-dependencies**\n\n```bash\npbt update\n```\n\n4. **Clean packages\' build & lock files**\n\n```bash\npbt clean [-p <package>]\n```\n\n- `-p`: specify the package we want to build, if empty build all packages.\n\n6. **Git clone a multi-repository project**\n\n```bash\npbt git clone --repo <repo_url>\n```\n\nClone a repository and check out all of its submodules to their correct branches that we were using last time.\n\n7. **Git update a multi-repository project**\n\n```bash\npbt git update\n```\n\nPull latest changes from the repository, and check out all of its submodules to their correct branches.\n',
    'author': 'Binh Vu',
    'author_email': 'binh@toan2.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/binh-vu/pbt',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from abc import abstractmethod
from contextlib import contextmanager
import glob
import os
from pathlib import Path
import shutil
from typing import Callable, Dict, List, Literal, Optional, Set, Union, cast

from loguru import logger
from pbt.config import PBTConfig
from pbt.misc import cache_method, exec
from pbt.package.manager.manager import PkgManager, build_cache
from pbt.package.package import DepConstraints, Package, PackageType
from tomlkit.api import loads, dumps
from pbt.diff import Diff, diff_db


class PythonPkgManager(PkgManager):
    """Package Managers for Python should inherit this class."""

    def __init__(self, cfg: PBTConfig, pkg_type: PackageType):
        super().__init__(cfg)
        self.managers: Dict[PackageType, PkgManager] = {pkg_type: self}
        self.fixed_version_pkgs = {"python"}
        self.passthrough_envs = [
            "PATH",
            "CC",
            "CXX",
            "LIBCLANG_PATH",
            "LD_LIBRARY_PATH",
            "C_INCLUDE_PATH",
            "CPLUS_INCLUDE_PATH",
        ]

    def set_package_managers(self, managers: Dict[PackageType, PkgManager]):
        """Set all package managers, which will be used to retrieve correct package manager
        for a particular python package.
        """
        self.managers = managers

    def get_fixed_version_pkgs(self):
        return self.fixed_version_pkgs

    def compute_pkg_hash(self, pkg: Package, target: Optional[str] = None) -> str:
        """Compute hash of the content of the package"""
        # build the package in release mode
        self.build(pkg, release=True, clean_dist=False)

        # obtain the hash of the package
        whl_path = self.wheel_path(pkg, target)
        assert whl_path is not None, whl_path
        output = exec(["pip", "hash", whl_path])[1]
        output = output[output.find("--hash=") + len("--hash=") :]
        assert output.startswith("sha256:")
        return output[len("sha256:") :]

    def clean(self, package: Package):
        """Remove previously installed dependencies and the environment where the package is installed, for a freshly start.
        In addition, this also cleans built artifacts.

        Args:
            package: The package to clean
        """
        path = (package.location / self.cfg.python_virtualenvs_path).absolute()
        if path.exists():
            shutil.rmtree(path)
        for eggdir in glob.glob(str(package.location / "*.egg-info")):
            shutil.rmtree(eggdir)
        shutil.rmtree(package.location / self.cfg.distribution_dir, ignore_errors=True)

    @cache_method()
    def venv_path(self, name: str, dir: Path) -> Path:
        """Get virtual environment path of the package, create it if doesn't exist.

        For a package manager that manages virtualenvs in different location, it should override this function.

        Arguments:
            name: name of the package
            dir: directory of the package
        """
        if not self.is_package_directory(dir):
            raise ValueError(
                f"{dir} seems to not contain any Python package. This won't return the right path. Please report this issue."
            )

        path = (dir / self.cfg.python_virtualenvs_path).absolute()
        if not path.exists():
            exec(f"{self.cfg.get_python_path()} -m venv {path}", env=["PATH"])

        if not path.exists():
            raise Exception(f"Environment path {path} doesn't exist")
        return Path(path)

    def tar_path(self, pkg: Package) -> Optional[Path]:
        tar_file = pkg.location / "dist" / f"{pkg.name}-{pkg.version}.tar.gz"
        if tar_file.exists():
            return tar_file
        return None

    def wheel_path(
        self,
        pkg: Package,
        target: Optional[str] = None,
    ) -> Optional[Path]:
        dist_dir = self.cfg.distribution_dir
        if target is None:
            whl_files = list((pkg.location / dist_dir).glob(f"{pkg.name}*.whl"))
            if len(whl_files) == 0:
                return None
            if len(whl_files) > 1:
                whl_files = list(
                    (pkg.location / dist_dir).glob(f"{pkg.name}-{pkg.version}*.whl")
                )
                if len(whl_files) != 1:
                    raise Exception(
                        "Multiple wheels found at: {}".format(
                            (pkg.location / dist_dir).absolute()
                        )
                    )

            return whl_files[0]

        whl_file = (
            pkg.location
            / dist_dir
            / f"{pkg.name.replace('-', '_')}-{pkg.version}-{target}.whl"
        )
        if whl_file.exists():
            return whl_file
        return None

    def pip_path(self, pkg: Package) -> Path:
        """Get Pip executable from the virtual environment of the package."""
        return self.venv_path(pkg.name, pkg.location) / "bin/pip"

    def python_path(self, pkg: Package) -> Path:
        """Get Python executable from the virtual environment of the package."""
        return self.venv_path(pkg.name, pkg.location) / "bin/python"

    def build(
        self,
        pkg: Package,
        skip_deps: Optional[List[str]] = None,
        additional_deps: Optional[Dict[str, DepConstraints]] = None,
        release: bool = True,
        clean_dist: bool = True,
    ):
        """Build the package. Support ignoring some dependencies to avoid installing and solving
        dependencies multiple times (not in the super interface as compiled languages will complain).
        """
        with build_cache() as built_pkgs:
            skip_deps = skip_deps or []
            additional_deps = additional_deps or {}

            build_ident = (pkg.name, pkg.version)
            build_opts = (tuple(skip_deps), tuple(sorted(additional_deps.keys())))
            if built_pkgs.get(build_ident, None) == build_opts:
                return

            with diff_db(pkg, self.cfg) as db:
                with self.change_dependencies(pkg, skip_deps, additional_deps):
                    diff = Diff.from_local(db, self, pkg)
                    if (
                        self.wheel_path(pkg) is not None
                        and self.tar_path(pkg) is not None
                    ):
                        if not diff.is_modified(db):
                            built_pkgs[build_ident] = build_opts
                            return

                    try:
                        dist_dir = pkg.location / self.cfg.distribution_dir
                        if clean_dist and (dist_dir).exists():
                            shutil.rmtree(str(dist_dir))
                        self._build_command(pkg, release)
                    finally:
                        diff.save(db)

                    built_pkgs[build_ident] = build_opts

    @abstractmethod
    def install(
        self,
        package: Package,
        include_dev: bool = False,
        skip_deps: Optional[List[str]] = None,
        additional_deps: Optional[Dict[str, DepConstraints]] = None,
        virtualenv: Optional[Path] = None,
    ):
        """Install the package, assuming the the specification is updated. Note that if the package is phantom, only the dependencies are installed.

        Note: don't expect this function will be able to find the local dependencies in the project
        as the manager relies on a package registry. For local dependencies, add them to `skip_deps` anddd use `install_dependency` to install them
        separately instead. Otherwise, you may get an error.

        Args:
            package: The package to install
            include_dev: Whether to install dev dependencies
            skip_deps: The dependencies to skip (usually the local ones we want to install in editable mode separately). This option is not guaranteed to compiled language
            additional_deps: The additional dependencies to install
            virtualenv: The virtual environment to install the package in. By default it is the virtual environment of the package
        """
        raise NotImplementedError()

    def install_dependency(
        self,
        pkg: Package,
        dependency: Package,
        skip_dep_deps: Optional[List[str]] = None,
    ):
        assert dependency.name not in self.cfg.phantom_packages, dependency.name
        assert self.managers is not None
        dep_manager = self.managers[dependency.type]
        assert isinstance(dep_manager, PythonPkgManager)

        dep_manager.install(
            dependency,
            skip_deps=skip_dep_deps or [],
            virtualenv=self.venv_path(pkg.name, pkg.location),
        )

    def publish(self, pkg: Package):
        self.build(pkg, release=True)
        exec("twine upload --skip-existing dist/*", cwd=pkg.location, env=["PATH"])

    def get_virtualenv_environment_variables(self, virtualenv: Path) -> dict:
        return {
            "VIRTUAL_ENV": str(virtualenv),
            "PATH": str(virtualenv / "bin") + os.pathsep + os.environ.get("PATH", ""),
        }

    @contextmanager  # type: ignore
    @abstractmethod
    def change_dependencies(
        self,
        pkg: Package,
        skip_deps: List[str],
        additional_deps: Dict[str, DepConstraints],
        disable: bool = False,
    ):
        """Temporary change the dependencies of the package.

        When skip_deps and additional_deps are both empty, this is a no-op.

        Args:
            pkg: The package to mask
            skip_deps: The dependencies to skip
            additional_deps: Additional dependencies to add
            disable: Whether to manually disable the mask or not
        """
        pass

    @abstractmethod
    def _build_command(self, pkg: Package, release: bool):
        """Run the build command for the package.

        Arguments:
            pkg: The package to build
            release: whether to build in release mode
        """
        pass


class Pep518PkgManager(PythonPkgManager):
    """A package manager for Python packages that use PEP 518 (pyproject.toml)

    Arguments:
        cfg: pbt's configuration
        backend: name of the building backend
    """

    def __init__(self, cfg: PBTConfig, pkg_type: PackageType, backend: str):
        super().__init__(cfg, pkg_type)
        self.backend = backend

        self.cache_pyprojects = {}

    def glob_query(self, root: Path) -> str:
        return str(root.absolute() / "**/pyproject.toml")

    def discover(
        self, root: Path, ignore_dirs: Set[Path], ignore_dirnames: Set[str]
    ) -> List[Path]:
        outs = []
        root = root.resolve()

        stack = [root]
        while len(stack) > 0:
            dir = stack.pop()
            if (
                dir.name.startswith(".")
                or dir.name in ignore_dirnames
                or dir in ignore_dirs
            ):
                continue
            if (dir / "pyproject.toml").exists():
                outs.append(dir)
            stack.extend([subdir for subdir in dir.iterdir() if subdir.is_dir()])

        return outs

    def is_package_directory(self, dir: Path) -> bool:
        if not (dir / "pyproject.toml").exists():
            return False

        return (
            self.parse_pyproject(dir / "pyproject.toml")["build-system"][
                "build-backend"
            ]
            == self.backend
        )

    def parse_pyproject(self, pyproject_file: Path) -> dict:
        """Parse project metadata from the directory.

        Arguments:
            pyproject_file: path to the pyproject.toml file

        Returns: dict, force the type to dictionary to silence the annoying type checker due to tomlkit
        """
        infile = str(pyproject_file.absolute())
        if infile not in self.cache_pyprojects:
            try:
                with open(infile, "r") as f:
                    self.cache_pyprojects[infile] = cast(dict, loads(f.read()))
            except:
                logger.error("Inavlid TOML file: {}", infile)
                raise
        return self.cache_pyprojects[infile]

    def update_pyproject(self, pyproject_file: Path, update_fn: Callable[[dict], bool]):
        """Update project metadata.

        Arguments:
            pyproject_file: path to the pyproject.toml file
            update_fn: function to update the document, return true if the document is modified
        """
        doc = self.parse_pyproject(pyproject_file)
        if update_fn(doc):
            success = False
            pyproject_file_backup = pyproject_file.parent / "pyproject.toml.backup"
            shutil.move(
                pyproject_file,
                pyproject_file_backup,
            )
            try:
                with open(pyproject_file, "w") as f:
                    f.write(dumps(doc))  # type: ignore
                del self.cache_pyprojects[str(pyproject_file.absolute())]
                success = True
            finally:
                if not success:
                    shutil.move(pyproject_file_backup, pyproject_file)
                else:
                    os.remove(pyproject_file_backup)



import os
from typing import Literal
import click
from loguru import logger
from pbt.config import PBTConfig

from pbt.vcs.git import Git


@click.command()
@click.option(
    "--repo",
    default="",
    help="Specify the multi-repository that we are working with. e.g., https://github.com/binh-vu/pbt",
)
@click.option("--cwd", default=".", help="Override current working directory")
@click.argument("command", type=click.Choice(["clone", "update", "push", "snapshot"]))
def git(repo: str, cwd: str, command: Literal["snapshot"]):
    """Execute Git command in a super-project"""
    cwd = os.path.abspath(cwd)

    if command == "clone":
        assert repo.endswith(".git"), f"Invalid repository: `{repo}`"

        # clone repository
        repo_dir = Git.clone_all(repo, cwd)

        # checkout the submodule to the correct branch
        for submodule in Git.find_submodules(repo_dir):
            logger.info("Checkout submodule {}", submodule)
            Git.auto_checkout_branch(submodule)
    elif command == "update":
        Git.pull(cwd, submodules=True)
        # checkout the submodule to the correct branch
        for submodule in Git.find_submodules(cwd):
            logger.info("Checkout submodule {}", submodule)
            Git.auto_checkout_branch(submodule)
    elif command == "push":
        pbt_cfg = PBTConfig.from_dir(cwd)
        cwd = str(pbt_cfg.cwd.absolute())
        Git.push(cwd)
        for submodule_dir in Git.find_submodules(cwd):
            Git.push(submodule_dir)
    elif command == "snapshot":
        pbt_cfg = PBTConfig.from_dir(cwd)
        cwd = str(pbt_cfg.cwd.absolute())

        for submodule_dir in Git.find_submodules(cwd):
            # get the current branch
            branch = Git.get_current_branch(submodule_dir)
            print(f"bash -c 'cd {submodule_dir}; git checkout {branch}; git pull'")
    else:
        raise Exception(f"Invalid command: {command}")



import re
import os
import shutil
import semver
from contextlib import contextmanager
from operator import attrgetter
from pathlib import Path
from typing import Any, Dict, List, Optional, Union, cast

from loguru import logger
from pbt.config import PBTConfig
from pbt.misc import ExecProcessError, NewEnvVar, cache_method, exec
from pbt.package.manager.python import Pep518PkgManager
from pbt.package.package import DepConstraint, DepConstraints, Package, PackageType
from tomlkit.api import document, dumps, inline_table, loads, nl, table
from tomlkit.items import Array, KeyType, Trivia, SingleKey


class Poetry(Pep518PkgManager):
    def __init__(self, cfg: PBTConfig) -> None:
        super().__init__(
            cfg, pkg_type=PackageType.Poetry, backend="poetry.core.masonry.api"
        )
        self.passthrough_envs = self.passthrough_envs + ["POETRY_VIRTUALENVS_PATH"]

    @cache_method()
    def get_poetry_version(self, path: str) -> str:
        """Get current poetry version"""
        out = exec("poetry --version", env={"PATH": path})
        m = re.match(r"^Poetry \(?version ([^)]+)\)?$", out[0].strip())
        if m is None:
            raise ValueError(
                "Cannot parse Poetry version. Please report if after updating Poetry, you still encounter this error."
            )
        return m.group(1)

    @contextmanager  # type: ignore
    def change_dependencies(
        self,
        pkg: Package,
        skip_deps: List[str],
        additional_deps: Dict[str, DepConstraints],
        disable: bool = False,
    ):
        """Temporary mask out selected dependencies of the package. This is usually used for installing the package.

        When skip_deps and additional_deps are both empty, this is a no-op.

        Args:
            pkg: The package to mask
            skip_deps: The dependencies to skip
            additional_deps: Additional dependencies to add
            disable: Whether to manually disable the mask or not
        """
        if disable or (len(skip_deps) + len(additional_deps)) == 0:
            yield None
            return

        with open(pkg.location / "pyproject.toml", "r") as f:
            doc = cast(dict, loads(f.read()))

        for dep in skip_deps:
            if dep in doc["tool"]["poetry"].get("dependencies", {}):
                r = doc["tool"]["poetry"]["dependencies"].remove(dep)
            elif dep in doc["tool"]["poetry"].get("dev-dependencies", {}):
                doc["tool"]["poetry"]["dev-dependencies"].remove(dep)

        for dep, specs in additional_deps.items():
            if dep not in doc["tool"]["poetry"]["dependencies"]:
                doc["tool"]["poetry"]["dependencies"][dep] = self.serialize_dep_specs(
                    specs
                )

        with open(self.cfg.pkg_cache_dir(pkg) / "pyproject.modified.toml", "w") as f:
            f.write(dumps(cast(Any, doc)))

        try:
            os.rename(
                pkg.location / "pyproject.toml",
                self.cfg.pkg_cache_dir(pkg) / "pyproject.origin.toml",
            )
            if (pkg.location / "poetry.lock").exists():
                os.rename(
                    pkg.location / "poetry.lock",
                    self.cfg.pkg_cache_dir(pkg) / "poetry.origin.lock",
                )
            shutil.copy(
                self.cfg.pkg_cache_dir(pkg) / "pyproject.modified.toml",
                pkg.location / "pyproject.toml",
            )
            if (self.cfg.pkg_cache_dir(pkg) / "poetry.modified.lock").exists():
                shutil.copy(
                    self.cfg.pkg_cache_dir(pkg) / "poetry.modified.lock",
                    pkg.location / "poetry.lock",
                )
            yield None
        finally:
            os.rename(
                self.cfg.pkg_cache_dir(pkg) / "pyproject.origin.toml",
                pkg.location / "pyproject.toml",
            )
            if (pkg.location / "poetry.lock").exists():
                os.rename(
                    pkg.location / "poetry.lock",
                    self.cfg.pkg_cache_dir(pkg) / "poetry.modified.lock",
                )
            if (self.cfg.pkg_cache_dir(pkg) / "poetry.origin.lock").exists():
                os.rename(
                    self.cfg.pkg_cache_dir(pkg) / "poetry.origin.lock",
                    pkg.location / "poetry.lock",
                )

    def load(self, dir: Path) -> Package:
        try:
            project_cfg = self.parse_pyproject(dir / "pyproject.toml")
            name = project_cfg["tool"]["poetry"]["name"]
            version = project_cfg["tool"]["poetry"]["version"]

            dependencies = {}
            dev_dependencies = {}

            for deps, cfg_key in [
                (dependencies, "dependencies"),
                (dev_dependencies, "dev-dependencies"),
            ]:
                for k, vs in project_cfg["tool"]["poetry"].get(cfg_key, {}).items():
                    if not isinstance(vs, list):
                        vs = [vs]
                    deps[k] = sorted(
                        (self.parse_dep_spec(v) for v in vs),
                        key=attrgetter("constraint"),
                    )

            # see https://python-poetry.org/docs/pyproject/#include-and-exclude
            # and https://python-poetry.org/docs/pyproject/#packages
            include = project_cfg["tool"]["poetry"].get("include", [])
            include.append(name)
            for pkg_cfg in project_cfg["tool"]["poetry"].get("packages", []):
                include.append(
                    os.path.join(pkg_cfg.get("from", ""), pkg_cfg["include"])
                )
            include = sorted(set(include))

            exclude = project_cfg["tool"]["poetry"].get("exclude", [])
        except:
            logger.error("Error while parsing configuration in {}", dir)
            raise

        return Package(
            name=name,
            version=version,
            dependencies=dependencies,
            dev_dependencies=dev_dependencies,
            type=PackageType.Poetry,
            location=dir,
            include=include,
            exclude=exclude,
        )

    def save(self, pkg: Package, poetry_file: Optional[Path] = None):
        poetry_file = poetry_file or pkg.location / "pyproject.toml"
        if not poetry_file.exists():
            with open(poetry_file, "w") as f:
                doc = document()

                tbl = table()
                tbl.add("name", pkg.name)
                tbl.add("version", pkg.version)
                tbl.add("description", "")
                tbl.add("authors", [])

                if len(pkg.exclude) > 0:
                    tbl.add("exclude", pkg.exclude)
                if sum(int(x != pkg.name) for x in pkg.include) > 0:
                    tbl.add("packages", [{"include": x} for x in pkg.include])

                doc.add(SingleKey("tool.poetry", t=KeyType.Bare), tbl)

                tbl = table()
                for dep, specs in pkg.dependencies.items():
                    tbl.add(dep, self.serialize_dep_specs(specs))
                doc.add(nl())
                doc.add(SingleKey("tool.poetry.dependencies", t=KeyType.Bare), tbl)

                tbl = table()
                for dep, specs in pkg.dev_dependencies.items():
                    tbl.add(dep, self.serialize_dep_specs(specs))
                doc.add(nl())
                doc.add(SingleKey("tool.poetry.dev-dependencies", t=KeyType.Bare), tbl)

                tbl = table()
                tbl.add("requires", ["poetry-core>=1.0.0"])
                tbl.add("build-backend", "poetry.core.masonry.api")
                doc.add(nl())
                doc.add("build-system", tbl)

                f.write(dumps(doc))
            return

        doc = self.parse_pyproject(poetry_file)

        def update_fn(doc: dict):
            is_modified = False

            if pkg.name != doc["tool"]["poetry"]["name"]:
                doc["tool"]["poetry"]["name"] = pkg.name
                is_modified = True

            if pkg.version != doc["tool"]["poetry"]["version"]:
                doc["tool"]["poetry"]["version"] = pkg.version
                is_modified = True

            if pkg.exclude != doc["tool"]["poetry"].get("exclude", []):
                doc["tool"]["poetry"]["exclude"] = pkg.exclude
                is_modified = True

            include = doc["tool"]["poetry"].get("include", [])
            include.append(pkg.name)
            for pkg_cfg in doc["tool"]["poetry"].get("packages", []):
                include.append(
                    os.path.join(pkg_cfg.get("from", ""), pkg_cfg["include"])
                )
            include = sorted(set(include))
            if pkg.include != include:
                doc["tool"]["poetry"]["packages"] = [
                    {"include": x} for x in pkg.include
                ]
                is_modified = True

            for dependencies, corr_key in [
                (pkg.dependencies, "dependencies"),
                (pkg.dev_dependencies, "dev-dependencies"),
            ]:
                dependencies: Dict[str, DepConstraints]
                for dep, specs in dependencies.items():
                    is_dep_modified = False
                    if dep not in doc["tool"]["poetry"][corr_key]:
                        is_dep_modified = True
                    else:
                        other_vers = doc["tool"]["poetry"][corr_key][dep]
                        if not isinstance(other_vers, list):
                            other_vers = [other_vers]
                        other_specs = [self.parse_dep_spec(v) for v in other_vers]
                        if specs != other_specs:
                            is_dep_modified = True

                    if is_dep_modified:
                        doc["tool"]["poetry"][corr_key][dep] = self.serialize_dep_specs(
                            specs
                        )
                        is_modified = True
            return is_modified

        self.update_pyproject(poetry_file, update_fn)

    def install(
        self,
        package: Package,
        include_dev: bool = False,
        skip_deps: Optional[List[str]] = None,
        additional_deps: Optional[Dict[str, DepConstraints]] = None,
        virtualenv: Optional[Path] = None,
    ):
        skip_deps = skip_deps or []
        additional_deps = additional_deps or {}

        if "python" in skip_deps:
            # we cannot skip python requirements
            skip_deps.remove("python")

        if virtualenv is None:
            virtualenv = self.venv_path(package.name, package.location)

        path = os.environ.get("PATH", "")
        env: List[Union[str, NewEnvVar]] = [
            x for x in self.passthrough_envs if x != "PATH"
        ]
        for k, v in self.get_virtualenv_environment_variables(virtualenv).items():
            env.append({"name": k, "value": v})
            if k == "PATH":
                path = v

        options = ""
        if not include_dev:
            ver = self.parse_version(self.get_poetry_version(path))

            if semver.VersionInfo(
                major=ver.major, minor=ver.minor
            ) < semver.VersionInfo(major=1, minor=2):
                options = "--no-dev"
            else:
                options = "--only=main"

        with self.change_dependencies(package, skip_deps, additional_deps):
            if (package.location / "poetry.lock").exists():
                try:
                    exec("poetry lock --check", cwd=package.location, env=env)
                except ExecProcessError:
                    logger.debug(
                        "poetry.lock is inconsistent with pyproject.toml, updating lock file..."
                    )
                    exec(
                        "poetry lock --no-update",
                        cwd=package.location,
                        capture_stdout=False,
                        env=env,
                    )

            exec(
                f"poetry install {options}",
                cwd=package.location,
                capture_stdout=False,
                env=env,
            )

    def _build_command(self, pkg: Package, release: bool):
        exec(
            "poetry build",
            cwd=pkg.location,
            env=self.passthrough_envs,
        )

    def parse_dep_spec(self, spec: Union[str, dict]) -> DepConstraint:
        if isinstance(spec, str):
            constraint = f"python=* markers="
            return DepConstraint(version_spec=spec, constraint=constraint)
        elif isinstance(spec, dict):
            if "version" not in spec:
                if "url" in spec:
                    # try to figure out the version from the URL if possible, otherwise, use 1.0.0
                    m = re.search(r"\d+.\d+.\d+", spec["url"])
                    if m is not None:
                        version_spec = f"=={m.group()}"
                    else:
                        version_spec = "==1.0.0"
                else:
                    raise NotImplementedError(
                        f"Not support specify dependency outside of Pypi yet. But found spec {spec}"
                    )
            else:
                version_spec = spec["version"]

            constraint = (
                f"python={spec.get('python', '*')} markers={spec.get('markers', '')}"
            )
            origin_spec = spec.copy()
            if "version" in origin_spec:
                origin_spec.pop("version")

            return DepConstraint(
                version_spec=version_spec,
                constraint=constraint,
                version_spec_field="version",
                origin_spec=origin_spec,
            )

    def serialize_dep_specs(self, specs: DepConstraints):
        items = []
        for spec in specs:
            if spec.origin_spec is None:
                item = spec.version_spec
            else:
                item = inline_table()
                item[cast(str, spec.version_spec_field)] = spec.version_spec
                for k, v in spec.origin_spec.items():
                    item[k] = v
            items.append(item)

        if len(items) == 1:
            return items[0]
        return Array(items, Trivia(), multiline=True)



import os
from contextlib import contextmanager
from dataclasses import dataclass
from pathlib import Path
import shutil
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Tuple, Union, cast

from loguru import logger
from pbt.config import PBTConfig
from pbt.misc import NewEnvVar, exec
from pbt.package.manager.manager import DepConstraints
from pbt.package.manager.python import Pep518PkgManager
from pbt.package.package import DepConstraint, Package, PackageType
from tomlkit.api import array, document, dumps, inline_table, loads, nl, table

if TYPE_CHECKING:
    from pbt.package.manager.poetry import Poetry


@dataclass
class MaturinPackage(Package):
    """Temporary class until we fix the optionals dependency"""

    optional_dep_name: Optional[str] = None


class Maturin(Pep518PkgManager):
    def __init__(self, cfg: PBTConfig) -> None:
        super().__init__(cfg, pkg_type=PackageType.Maturin, backend="maturin")

    def load(self, dir: Path) -> MaturinPackage:
        try:
            with open(dir / "pyproject.toml", "r") as f:
                # just force the type to dictionary to silence the annoying type checker
                project_cfg = cast(dict, loads(f.read()))

                name = project_cfg["project"]["name"]
                version = project_cfg["project"]["version"]

                dependencies = {}
                for item in project_cfg["project"]["dependencies"]:
                    k, specs = self.parse_dep_spec(item)
                    dependencies[k] = specs

                dev_dependencies = {}
                opt = self.get_optional_dependency_name(project_cfg)
                if opt is not None:
                    for item in project_cfg["project"]["optional-dependencies"][opt]:
                        k, specs = self.parse_dep_spec(item)
                        dev_dependencies[k] = specs

                # not supported yet in pep-621
                # https://peps.python.org/pep-0621/#specify-files-to-include-when-building
                include = []
                exclude = []
                if "tool" in project_cfg and "maturin" in project_cfg["tool"]:
                    if "include" in project_cfg["tool"]["maturin"]:
                        include = project_cfg["tool"]["maturin"]["include"]
                    if "exclude" in project_cfg["tool"]["maturin"]:
                        exclude = project_cfg["tool"]["maturin"]["exclude"]
        except:
            logger.error("Error while parsing configuration in {}", dir)
            raise

        return MaturinPackage(
            name=name,
            version=version,
            dependencies=dependencies,
            dev_dependencies=dev_dependencies,
            type=PackageType.Maturin,
            location=dir,
            include=include,
            exclude=exclude,
            optional_dep_name=opt,
        )

    def save(self, pkg: Package):
        tomlfile = pkg.location / "pyproject.toml"
        if not tomlfile.exists():
            raise NotImplementedError(
                "Don't support creating new pyproject.toml file yet"
            )

        old_pkg = self.load(pkg.location)

        with open(tomlfile, "r") as f:
            doc = cast(Any, loads(f.read()))
            is_modified = False

            if pkg.name != doc["project"]["name"]:
                doc["project"]["name"] = pkg.name
                is_modified = True

            if pkg.version != doc["project"]["version"]:
                doc["project"]["version"] = pkg.version
                is_modified = True

            for dependencies, old_dependencies, corr_key in [
                (pkg.dependencies, old_pkg.dependencies, "dependencies"),
                (
                    pkg.dev_dependencies,
                    old_pkg.dev_dependencies,
                    "optional-dependencies",
                ),
            ]:
                dependencies: Dict[str, DepConstraints]
                is_dep_modified = False
                for dep, specs in dependencies.items():
                    if dep not in old_dependencies or old_dependencies[dep] != specs:
                        is_dep_modified = True
                        break
                if is_dep_modified:
                    is_modified = True
                    lst = array(
                        [
                            dep + " " + self.serialize_dep_specs(specs)
                            for dep, specs in dependencies.items()
                        ]  # type: ignore
                    )
                    lst.multiline(True)
                    if corr_key == "dependencies":
                        doc["project"][corr_key] = lst
                    else:
                        opt = self.get_optional_dependency_name(doc)
                        assert (
                            opt is not None
                        ), "The dep is modified, we should have optional-dependencies"
                        doc["project"][corr_key][opt] = lst

        if is_modified:
            success = False
            os.rename(
                pkg.location / "pyproject.toml",
                pkg.location / "pyproject.toml.backup",
            )
            try:
                with open(tomlfile, "w") as f:
                    f.write(dumps(doc))
                success = True
            finally:
                if not success:
                    os.rename(
                        pkg.location / "pyproject.toml.backup",
                        pkg.location / "pyproject.toml",
                    )
                else:
                    os.remove(pkg.location / "pyproject.toml.backup")

    def parse_dep_spec(self, v: str) -> Tuple[str, DepConstraints]:
        """Parse a dependency specification.

        It does not support PEP508 but only a simple syntax: `<name> <version_rule>`.
        Note: the space is important.
        """
        name, version = v.split(" ", 1)
        # do it here to make sure we can parse this version
        self.parse_version_spec(version)
        constraint = f"python=* markers="
        return name, [DepConstraint(version_spec=version, constraint=constraint)]

    def serialize_dep_specs(self, specs: DepConstraints) -> str:
        # not implement all cases yet
        assert len(specs) == 1
        (spec,) = specs
        assert spec.origin_spec is None
        version_spec = self.parse_version_spec(spec.version_spec)
        return version_spec.to_pep508_string()

    def clean(self, pkg: Package):
        super().clean(pkg)
        # run cargo clean to clean up rust build artifacts
        exec(
            "cargo clean",
            cwd=pkg.location,
            env=self.passthrough_envs,
        )

    def install(
        self,
        pkg: MaturinPackage,
        include_dev: bool = False,
        skip_deps: Optional[List[str]] = None,
        additional_deps: Optional[Dict[str, DepConstraints]] = None,
        virtualenv: Optional[Path] = None,
    ):
        skip_deps = skip_deps or []
        additional_deps = additional_deps or {}
        options = ""
        if include_dev and pkg.optional_dep_name is not None:
            options += " --extras=" + pkg.optional_dep_name

        if virtualenv is None:
            virtualenv = self.venv_path(pkg.name, pkg.location)

        # set the virtual environment which the package will be installed to
        env: List[Union[str, NewEnvVar]] = [
            x for x in self.passthrough_envs if x != "PATH"
        ]
        for k, v in self.get_virtualenv_environment_variables(virtualenv).items():
            env.append({"name": k, "value": v})

        with self.change_dependencies(pkg, skip_deps, additional_deps):
            exec(f"maturin develop -r {options}", cwd=pkg.location, env=env)

    def _build_command(self, pkg: Package, release: bool):
        cmd: List[Union[str, Path]] = [
            "maturin",
            "build",
        ]
        if release:
            cmd.append("-r")
        cmd.extend(["-o", (pkg.location / self.cfg.distribution_dir).absolute()])

        exec(
            cmd,
            cwd=pkg.location,
            env=self.passthrough_envs,
        )

    def get_optional_dependency_name(self, doc: dict) -> Optional[str]:
        """Get the name of the optional dependency (i.e., extras) in the pyproject.toml."""
        opts = list(doc["project"]["optional-dependencies"].keys())
        if len(opts) == 1:
            return opts[0]

        if len(opts) > 1:
            raise NotImplementedError(
                "Haven't support multiple options-dependencies yet"
            )

        return None

    @contextmanager  # type: ignore
    def change_dependencies(
        self,
        pkg: Package,
        skip_deps: List[str],
        additional_deps: Dict[str, DepConstraints],
        disable: bool = False,
    ):
        """Temporary mask out selected dependencies of the package. This is usually used for installing the package.

        When skip_deps and additional_deps are both empty, this is a no-op.

        Args:
            pkg: The package to mask
            skip_deps: The dependencies to skip
            additional_deps: Additional dependencies to add
            disable: Whether to manually disable the mask or not
        """
        if disable or (len(skip_deps) + len(additional_deps)) == 0:
            yield None
            return

        with open(pkg.location / "pyproject.toml", "r") as f:
            doc = cast(dict, loads(f.read()))

        dep_name2lineno = {}
        optdep_name2lineno = {}
        for i, item in enumerate(doc["project"]["dependencies"]):
            name, specs = self.parse_dep_spec(item)
            dep_name2lineno[name] = i

        removed_lines = {
            dep_name2lineno[dep] for dep in skip_deps if dep in dep_name2lineno
        }
        doc["project"]["dependencies"] = [
            line
            for i, line in enumerate(doc["project"]["dependencies"])
            if i not in removed_lines
        ]

        opt = self.get_optional_dependency_name(doc)
        if opt is not None:
            for i, item in enumerate(doc["project"]["optional-dependencies"][opt]):
                name, specs = self.parse_dep_spec(item)
                optdep_name2lineno[name] = i

            removed_lines = {
                optdep_name2lineno[dep]
                for dep in skip_deps
                if dep in optdep_name2lineno
            }
            doc["project"]["optional-dependencies"][opt] = [
                line
                for i, line in enumerate(doc["project"]["optional-dependencies"][opt])
                if i not in removed_lines
            ]

        for dep, specs in additional_deps.items():
            if dep not in dep_name2lineno:
                doc["project"]["dependencies"].append(
                    dep + " " + self.serialize_dep_specs(specs)
                )

        with open(self.cfg.pkg_cache_dir(pkg) / "pyproject.modified.toml", "w") as f:
            f.write(dumps(cast(Any, doc)))

        try:
            os.rename(
                pkg.location / "pyproject.toml",
                self.cfg.pkg_cache_dir(pkg) / "pyproject.origin.toml",
            )
            shutil.copy(
                self.cfg.pkg_cache_dir(pkg) / "pyproject.modified.toml",
                pkg.location / "pyproject.toml",
            )
            yield None
        finally:
            os.rename(
                self.cfg.pkg_cache_dir(pkg) / "pyproject.origin.toml",
                pkg.location / "pyproject.toml",
            )
